"""
Performs an end-to-end test where we verify that the data
output by the client command line interface is equal to
the values we expect using the test dataset.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import datetime
import json
import unittest
import freezegun

import ga4gh.client.client as client
import ga4gh.server.backend as backend
import ga4gh.client.cli as cli_client
import ga4gh.server.datarepo as datarepo
import ga4gh.common.utils as utils
import tests.paths as paths
import server

import ga4gh.schemas.protocol as protocol


def setUpModule():
    global moduleTestServer
    moduleTestServer = server.Ga4ghServerForTestingDataSource(
        paths.testDataRepo)
    moduleTestServer.start()


def tearDownModule():
    moduleTestServer.shutdown()


@freezegun.freeze_time(datetime.datetime.now())
class TestClientOutput(unittest.TestCase):
    """
    Base class for client output tests
    """
    def setUp(self):
        self._maxDiff = None
        repoPath = paths.testDataRepo
        self._dataUrl = moduleTestServer.getUrl()
        dataRepository = datarepo.SqlDataRepository(repoPath)
        dataRepository.open(datarepo.MODE_READ)
        self._backend = backend.Backend(dataRepository)
        self._client = client.LocalClient(self._backend)
        # TODO probably could use a cache of objects, so we don't
        # need to keep re-fetching them

    def captureCliOutput(self, command, arguments, outputFormat):
        clientCommand = "{} {} {} -O {}".format(
            command, self._dataUrl, arguments, outputFormat)
        stdout, stderr = utils.captureOutput(
            cli_client.client_main, clientCommand.split())
        self.assertEqual(len(stderr), 0)
        return stdout


class TestClientFasta(TestClientOutput):
    """
    Tests client FASTA output
    """
    def captureFastaOutput(self, command, arguments=""):
        stdout = self.captureCliOutput(command, arguments, "fasta")
        lines = stdout.split()
        return lines

    def testListReferenceBases(self):
        referenceSetIterator = self._client.search_reference_sets()
        referenceSet = next(referenceSetIterator)
        referencesIterator = self._client.search_references(referenceSet.id)
        reference = next(referencesIterator)
        start = 1
        end = 5
        lines = self.captureFastaOutput(
            "references-list-bases --start {} --end {}".format(start, end),
            reference.id)
        self.assertEqual(
            lines[0], ">{}:{}-{}".format(reference.id, start, end))
        cliBases = ''.join(lines[1:])
        bases = self._client.list_reference_bases(reference.id, start, end)
        self.assertEqual(cliBases, bases)


class TestClientJson(TestClientOutput):
    """
    Tests that the JSON output by the client on the command line for
    various options is equal to the values we find using the Python
    client API.
    """
    def captureJsonOutput(self, command, arguments=""):
        """
        Runs the specified command add the JSON output option and
        returns the result as a list of JSON parsed dictionaries.
        """
        stdout = self.captureCliOutput(command, arguments, "json")
        cliOutput = []
        for line in stdout.splitlines():
            try:
                cliOutput.append(json.loads(line))
            except ValueError, e:
                raise Exception((e, line, stdout, command, arguments))
        return cliOutput

    def verifyParsedOutputsEqual(
            self, clientIterator, cliCommand, cliArguments="",
            scrubFunc=None):
        """
        Verify that the parsed JSON of all the objects in the specified
        client iterator are equal to the parsed JSON from the specified
        CLI command.
        """
        cliOutput = self.captureJsonOutput(cliCommand, cliArguments)
        clientOutput = [
            protocol.toJsonDict(gObj) for gObj in clientIterator]
        if scrubFunc is not None:
            # this is a hack to deal with the issue of the timestamps on
            # objects being different
            scrubFunc(cliOutput, clientOutput)
        self.assertEqual(clientOutput, cliOutput)
        return len(clientOutput)

    def _scrubReadGroup(self, cliReadGroup, clientReadGroup):
        cliReadGroup['created'] = clientReadGroup['created']
        cliReadGroup['updated'] = clientReadGroup['updated']
        cliReadGroup['experiment']['messageUpdateTime'] = \
            clientReadGroup['experiment']['messageUpdateTime']
        cliReadGroup['experiment']['messageCreateTime'] = \
            clientReadGroup['experiment']['messageCreateTime']

    def scrubReadGroup(self, cliOutput, clientOutput):
        for cliReadGroup, clientReadGroup in utils.zipLists(
                cliOutput, clientOutput):
            self._scrubReadGroup(cliReadGroup, clientReadGroup)

    def scrubReadGroupSet(self, cliOutput, clientOutput):
        for cliReadGroupSet, clientReadGroupSet in utils.zipLists(
                cliOutput, clientOutput):
            for cliReadGroup, clientReadGroup in utils.zipLists(
                    cliReadGroupSet['readGroups'],
                    clientReadGroupSet['readGroups']):
                self._scrubReadGroup(cliReadGroup, clientReadGroup)

    def verifyParsedOutputsEqualReadGroup(
            self, clientIterator, cliCommand, cliArguments=""):
        cliOutput = self.captureJsonOutput(cliCommand, cliArguments)
        clientOutput = [
            protocol.toJsonDict(gObj) for gObj in clientIterator]
        self.assertEqual(clientOutput, cliOutput)
        return len(clientOutput)

    def testGetCallSet(self):
        for dataset in self._client.search_datasets():
            for variantSet in self._client.search_variant_sets(dataset.id):
                for callSet in self._client.search_call_sets(variantSet.id):
                    self.verifyParsedOutputsEqual(
                        [callSet], "callsets-get", callSet.id)

    def testGetDataset(self):
        for dataset in self._client.search_datasets():
            self.verifyParsedOutputsEqual(
                [dataset], "datasets-get", dataset.id)

    def testGetReadGroup(self):
        for dataset in self._client.search_datasets():
            for readGroupSet in self._client.search_read_group_sets(
                    dataset.id):
                for readGroup in readGroupSet.read_groups:
                    self.verifyParsedOutputsEqual(
                        [readGroup], "readgroups-get", readGroup.id,
                        scrubFunc=self.scrubReadGroup)

    def testGetReadGroupSet(self):
        for dataset in self._client.search_datasets():
            for readGroupSet in self._client.search_read_group_sets(
                    dataset.id):
                self.verifyParsedOutputsEqual(
                    [readGroupSet], "readgroupsets-get", readGroupSet.id,
                    scrubFunc=self.scrubReadGroupSet)

    def testGetReference(self):
        for referenceSet in self._client.search_reference_sets():
            for reference in self._client.search_references(
                    referenceSet.id):
                self.verifyParsedOutputsEqual(
                    [reference], "references-get", reference.id)

    def testGetReferenceSet(self):
        for referenceSet in self._client.search_reference_sets():
            self.verifyParsedOutputsEqual(
                [referenceSet], "referencesets-get", referenceSet.id)

    @unittest.skip("TODO: clarify semantics of callsets and fix")
    def testGetVariant(self):
        test_executed = 0
        start = 0
        end = 1000
        referenceName = "1"
        for dataset in self._client.search_datasets():
            for variantSet in self._client.search_variant_sets(dataset.id):
                variants = self._client.search_variants(
                    variantSet.id, start=start, end=end,
                    reference_name=referenceName)
                for variant in variants:
                    test_executed += self.verifyParsedOutputsEqual(
                        [variant], "variants-get", variant.id)
        self.assertGreater(test_executed, 0)

    def testGetVariantAnnotationSet(self):
        # TODO this doesn't actually test get_variant_annotation_set
        test_executed = 0
        for dataset in self._client.search_datasets():
            for variantSet in self._client.search_variant_sets(dataset.id):
                for annSet in self._client.search_variant_annotation_sets(
                        variantSet.id):
                    test_executed += self.verifyParsedOutputsEqual(
                        [annSet], "variantannotationsets-get", annSet.id)
        self.assertGreater(test_executed, 0)

    def testGetVariantSet(self):
        for dataset in self._client.search_datasets():
            for variantSet in self._client.search_variant_sets(dataset.id):
                self.verifyParsedOutputsEqual(
                    [variantSet], "variantsets-get", variantSet.id)

    def testSearchCallSets(self):
        for dataset in self._client.search_datasets():
            for variantSet in self._client.search_variant_sets(dataset.id):
                iterator = self._client.search_call_sets(variantSet.id)
                args = "--variantSetId {}".format(variantSet.id)
                self.verifyParsedOutputsEqual(
                    iterator, "callsets-search", args)

    def testGetBiosamples(self):
        for dataset in self._client.search_datasets():
            for biosample in self._client.search_biosamples(dataset.id):
                self.verifyParsedOutputsEqual(
                    [biosample], "biosamples-get", biosample.id)

    def testGetIndividuals(self):
        for dataset in self._client.search_datasets():
            for individual in self._client.search_individuals(dataset.id):
                self.verifyParsedOutputsEqual(
                    [individual], "individuals-get", individual.id)

    def testSearchDatasets(self):
        iterator = self._client.search_datasets()
        self.verifyParsedOutputsEqual(iterator, "datasets-search")

    def testSearchReadGroupSets(self):
        for dataset in self._client.search_datasets():
            iterator = self._client.search_read_group_sets(dataset.id)
            self.verifyParsedOutputsEqual(
                iterator, "readgroupsets-search",
                "--datasetId {}".format(dataset.id),
                scrubFunc=self.scrubReadGroupSet)
            for readGroupSet in iterator:
                bioIterator = self._client.search_read_group_sets(
                    dataset.id,
                    name=readGroupSet.name,
                    biosample_id=readGroupSet.biosampleId)
                self.verifyParsedOutputsEqual(
                    bioIterator, "readgroupsets-search",
                    "--datasetId {} --name {} --biosampleId {}".format(
                        dataset.id,
                        readGroupSet.name,
                        readGroupSet.biosampleId),
                    scrubFunc=self.scrubReadGroupSet)

    def testSearchReads(self):
        test_executed = 0
        start = 0
        end = 1000000
        for dataset in self._client.search_datasets():
            for readGroupSet in self._client.search_read_group_sets(
                    dataset.id):
                for readGroup in readGroupSet.read_groups:
                    reference = self._client.search_references(
                        reference_set_id=readGroup.reference_set_id).next()
                    referenceId = reference.id
                    iterator = self._client.search_reads(
                        [readGroup.id], reference_id=referenceId,
                        start=start, end=end)
                    args = "--start {} --end {} --readGroupIds {}\
                    --referenceId {}".format(
                        start, end, readGroup.id, referenceId)
                    test_executed += self.verifyParsedOutputsEqual(
                        iterator, "reads-search", args)
        self.assertGreater(test_executed, 0)

    def testSearchReferenceSets(self):
        iterator = self._client.search_reference_sets()
        self.verifyParsedOutputsEqual(iterator, "referencesets-search")

    def testSearchReferences(self):
        for referenceSet in self._client.search_reference_sets():
            iterator = self._client.search_references(
                reference_set_id=referenceSet.id)
            args = "--referenceSetId={}".format(referenceSet.id)
            self.verifyParsedOutputsEqual(iterator, "references-search", args)

    def testSearchVariantSets(self):
        for dataset in self._client.search_datasets():
            iterator = self._client.search_variant_sets(dataset.id)
            self.verifyParsedOutputsEqual(iterator, "variantsets-search")

    def testSearchVariants(self):
        test_executed = 0
        start = 0
        end = 1000
        referenceName = "1"
        for dataset in self._client.search_datasets():
            for variantSet in self._client.search_variant_sets(dataset.id):
                iterator = self._client.search_variants(
                    variantSet.id, start=start, end=end,
                    reference_name=referenceName, call_set_ids=[])
                args = "--variantSetId {} --start {} --end {} -r {}".format(
                    variantSet.id, start, end, referenceName)
                test_executed += self.verifyParsedOutputsEqual(
                    iterator, "variants-search", args)
        self.assertGreater(test_executed, 0)

    def testSearchBiosamples(self):
        for dataset in self._client.search_datasets():
            iterator = self._client.search_biosamples(dataset.id)
            self.verifyParsedOutputsEqual(iterator, "biosamples-search")
        for dataset in self._client.search_datasets():
            for biosample in self._client.search_biosamples(dataset.id):
                iterator = self._client.search_biosamples(
                    dataset.id, biosample.name)
                self.verifyParsedOutputsEqual(
                    iterator,
                    "biosamples-search",
                    "--datasetId {} --name {}".format(
                        dataset.id,
                        biosample.name))
                if biosample.individualId:
                    iterator = self._client.search_biosamples(
                        dataset.id, individual_id=biosample.individualId)
                    self.verifyParsedOutputsEqual(
                        iterator,
                        "biosamples-search",
                        "--datasetId {} --individualId {}".format(
                            dataset.id,
                            biosample.individualId))

    def testSearchIndividuals(self):
        for dataset in self._client.search_datasets():
            iterator = self._client.search_individuals(dataset.id)
            self.verifyParsedOutputsEqual(iterator, "individuals-search")
        for dataset in self._client.search_datasets():
            for individual in self._client.search_individuals(dataset.id):
                iterator = self._client.search_individuals(
                    dataset.id, individual.name)
                self.verifyParsedOutputsEqual(
                    iterator, "individuals-search", "--name {}".format(
                        individual.name))

    def testSearchVariantAnnotationSets(self):
        for dataset in self._client.search_datasets():
            for variantSet in self._client.search_variant_sets(dataset.id):
                iterator = self._client.search_variant_annotation_sets(
                    variantSet.id)
                args = "{}".format(variantSet.id)
                self.verifyParsedOutputsEqual(
                    iterator, "variantannotationsets-search", args)

    def testSearchVariantAnnotations(self):
        test_executed = 0
        start = 0
        end = 10000000
        referenceName = "1"
        for dataset in self._client.search_datasets():
            for variantSet in self._client.search_variant_sets(dataset.id):
                searchIterator = self._client.search_variant_annotation_sets(
                    variantSet.id)
                for variantAnnotationSet in searchIterator:
                    iterator = self._client.search_variant_annotations(
                        variantAnnotationSet.id,
                        start=start,
                        end=end,
                        reference_name=referenceName)
                    args = ("--variantAnnotationSetId {}"
                            " --start {} --end {} -r {}").format(
                        variantAnnotationSet.id, start, end, referenceName)
                    test_executed += self.verifyParsedOutputsEqual(
                        iterator, "variantannotations-search", args)
        self.assertGreater(test_executed, 0)

    def testGetFeatures(self):
        for dataset in self._client.search_datasets():
            datasetId = dataset.id
            for featureSet in self._client.search_feature_sets(datasetId):
                for feature in self._client.search_features(
                        featureSet.id):
                    self.verifyParsedOutputsEqual(
                        [feature], "features-get", feature.id)
                    break  # this test takes too long otherwise

    def testSearchFeatures(self):
        for dataset in self._client.search_datasets():
            datasetId = dataset.id
            for featureSet in self._client.search_feature_sets(datasetId):
                iterator = self._client.search_features(featureSet.id)
                self.verifyParsedOutputsEqual(
                    iterator, "features-search",
                    "--featureSetId {}".format(featureSet.id))

    def testGetFeatureSets(self):
        for dataset in self._client.search_datasets():
            datasetId = dataset.id
            for featureSet in self._client.search_feature_sets(datasetId):
                self.verifyParsedOutputsEqual(
                    [featureSet], "featuresets-get", featureSet.id)

    def testSearchFeatureSets(self):
        for dataset in self._client.search_datasets():
            iterator = self._client.search_feature_sets(dataset.id)
            self.verifyParsedOutputsEqual(
                iterator, "featuresets-search",
                "--datasetId {}".format(dataset.id))

    def testSearchContinuous(self):
        for dataset in self._client.search_datasets():
            datasetId = dataset.id
            for continuousSet in self._client.search_continuous_sets(
                                                datasetId):
                iterator = self._client.search_continuous(
                                continuousSet.id, 'chr19', 49305897, 49306090)
                self.verifyParsedOutputsEqual(
                    iterator, "continuous-search",
                    "--continuousSetId {} --referenceName {}"
                    " --start {} --end {}".format(
                        continuousSet.id, 'chr19', 49305897, 49306090))

    def testGetContinuousSets(self):
        for dataset in self._client.search_datasets():
            datasetId = dataset.id
            for continuousSet in self._client.search_continuous_sets(
                                                datasetId):
                self.verifyParsedOutputsEqual(
                    [continuousSet], "continuoussets-get", continuousSet.id)

    def testSearchContinuousSets(self):
        for dataset in self._client.search_datasets():
            iterator = self._client.search_continuous_sets(dataset.id)
            self.verifyParsedOutputsEqual(
                iterator, "continuoussets-search",
                "--datasetId {}".format(dataset.id))

    def testSearchGenotypePhenotype(self):
        phenotype_id = "http://ohsu.edu/cgd/87795e43"
        test_executed = 0
        for dataset in self._client.search_datasets():
            # pas = phenotype_association_set
            for pas in \
              self._client.search_phenotype_association_sets(dataset.id):
                    iterator = self._client.search_genotype_phenotype(
                        phenotype_association_set_id=pas.id,
                        phenotype_ids=[phenotype_id])
                    args = (
                        "--phenotype_association_set_id {}"
                        " --phenotype_ids {} ").format(
                        pas.id, phenotype_id)
                    test_executed += self.verifyParsedOutputsEqual(
                        iterator, "genotypephenotype-search", args)
        self.assertGreater(test_executed, 0)

    def testSearchPhenotype(self):
        phenotype_id = "http://ohsu.edu/cgd/87795e43"
        test_executed = 0
        for dataset in self._client.search_datasets():
            # pas = phenotype_association_set
            for pas in \
              self._client.search_phenotype_association_sets(dataset.id):
                    iterator = self._client.search_phenotype(
                        phenotype_association_set_id=pas.id,
                        phenotype_id=phenotype_id)
                    args = (
                        "--phenotype_association_set_id {}"
                        " --phenotype_id {} ").format(
                        pas.id, phenotype_id)
                    test_executed += self.verifyParsedOutputsEqual(
                        iterator, "phenotype-search", args)
        self.assertGreater(test_executed, 0)

    def testSearchSearchPhenotypeAssociationSets(self):
        test_executed = 0
        for dataset in self._client.search_datasets():
            iterator = self._client.search_phenotype_association_sets(
                dataset_id=dataset.id)
            args = "--datasetId {}".format(dataset.id)
            test_executed += self.verifyParsedOutputsEqual(
                iterator, "phenotypeassociationsets-search", args)
        self.assertGreater(test_executed, 0)

    def testSearchExpressionLevels(self):
        for dataset in self._client.search_datasets():
            for rnaQuantificationSet in \
                    self._client.search_rna_quantification_sets(dataset.id):
                for rnaQuantification in \
                        self._client.search_rna_quantifications(
                                rnaQuantificationSet.id):
                    iterator = self._client.search_expression_levels(
                        rnaQuantification.id)
                    cliString = (
                        "expressionlevels-search "
                        "--rnaQuantificationId {}".format(
                            rnaQuantification.id))
                    self.verifyParsedOutputsEqual(iterator, cliString)

    def testSearchRnaQuantifications(self):
        for dataset in self._client.search_datasets():
            for rnaQuantificationSet in \
                    self._client.search_rna_quantification_sets(dataset.id):
                iterator = self._client.search_rna_quantifications(
                    rnaQuantificationSet.id)
                cliString = (
                    "rnaquantifications-search "
                    "--rnaQuantificationSetId {}".format(
                        rnaQuantificationSet.id))
                self.verifyParsedOutputsEqual(iterator, cliString)

    def testSearchRnaQuantificationSets(self):
        for dataset in self._client.search_datasets():
            iterator = self._client.search_rna_quantification_sets(dataset.id)
            cliString = (
                "rnaquantificationsets-search --datasetId {}".format(
                    dataset.id))
            self.verifyParsedOutputsEqual(iterator, cliString)

    def testGetExpressionLevel(self):
        for dataset in self._client.search_datasets():
            for rnaQuantificationSet in \
                    self._client.search_rna_quantification_sets(dataset.id):
                for rnaQuantification in \
                        self._client.search_rna_quantifications(
                            rnaQuantificationSet.id):
                            for expressionLevel in \
                                    self._client.search_expression_levels(
                                        rnaQuantification.id):
                                self.verifyParsedOutputsEqual(
                                    [expressionLevel],
                                    "expressionlevels-get",
                                    expressionLevel.id)

    def testGetRnaQuantification(self):
        for dataset in self._client.search_datasets():
            for rnaQuantificationSet in \
                    self._client.search_rna_quantification_sets(dataset.id):
                for rnaQuantification in \
                        self._client.search_rna_quantifications(
                            rnaQuantificationSet.id):
                    self.verifyParsedOutputsEqual(
                        [rnaQuantification],
                        "rnaquantifications-get",
                        rnaQuantification.id)

    def testGetRnaQuantificationSet(self):
        for dataset in self._client.search_datasets():
            for rnaQuantificationSet in \
                    self._client.search_rna_quantification_sets(dataset.id):
                self.verifyParsedOutputsEqual(
                    [rnaQuantificationSet],
                    "rnaquantificationsets-get",
                    rnaQuantificationSet.id)

    def testListPeers(self):
        iterator = self._client.list_peers()
        cliString = "list-peers"
        self.verifyParsedOutputsEqual(iterator, cliString)

    def testInfo(self):
        info = self._client.get_info()
        cliString = "get-info"
        self.verifyParsedOutputsEqual([info], cliString)

    def testAnnounce(self):
        url = "http://1kgenomes.ga4gh.org"
        response = self._client.announce(url)
        cliString = "announce"
        self.verifyParsedOutputsEqual([response], cliString, url)



#!/usr/bin/env python
import argparse
from functools import wraps, partial
import json
import mimetypes
import os
import urllib
import urlparse
import errno

import cherrypy
from cherrypy import wsgiserver
from cherrypy.wsgiserver import ssl_pyopenssl
from cherrypy.wsgiserver.wsgiserver2 import WSGIPathInfoDispatcher
from jinja2.environment import Environment
from jinja2.loaders import FileSystemLoader
import yaml

from oic.oauth2 import rndstr
from oic.oic.provider import Provider, AuthorizationEndpoint, TokenEndpoint, \
    UserinfoEndpoint, RegistrationEndpoint, EndSessionEndpoint
from oic.utils.authn.authn_context import AuthnBroker
from oic.utils.authn.client import verify_client
from oic.utils.authz import AuthzHandling
from oic.utils.http_util import Redirect, SeeOther, Response, BadRequest, \
    get_or_post, get_post, NotFound
from oic.utils.keyio import keyjar_init
from oic.utils.sdb import SessionDB
from oic.utils.userinfo import UserInfo
from oic.utils.webfinger import OIC_ISSUER, WebFinger
from provider.authn import make_cls_from_name


def VerifierMiddleware(verifier):
    """Common wrapper for the authentication modules.
        * Parses the request before passing it on to the authentication module.
        * Sets 'pyoidc' cookie if authentication succeeds.
        * Redirects the user to complete the authentication.
        * Allows the user to retry authentication if it fails.
    :param verifier: authentication module
    """

    @wraps(verifier.verify)
    def wrapper(environ, start_response):
        data = get_post(environ)
        kwargs = dict(urlparse.parse_qsl(data))
        kwargs["state"] = json.loads(urllib.unquote(kwargs["state"]))
        val, completed = verifier.verify(**kwargs)
        if not completed:
            return val(environ, start_response)
        if val:
            set_cookie, cookie_value = verifier.create_cookie(val, "auth")
            cookie_value += "; path=/"

            url = "{base_url}?{query_string}".format(
                base_url="/authorization",
                query_string=kwargs["state"]["query"])
            response = SeeOther(url, headers=[(set_cookie, cookie_value)])
            return response(environ, start_response)
        else:  # Unsuccessful authentication
            url = "{base_url}?{query_string}".format(
                base_url="/authorization",
                query_string=kwargs["state"]["query"])
            response = SeeOther(url)
            return response(environ, start_response)

    return wrapper


def pyoidcMiddleware(func):
    """Common wrapper for the underlying pyoidc library functions.
    Reads GET params and POST data before passing it on the library and
    converts the response from oic.utils.http_util to wsgi.
    :param func: underlying library function
    """

    def wrapper(environ, start_response):
        data = get_or_post(environ)
        cookies = environ.get("HTTP_COOKIE", "")
        resp = func(request=data, cookie=cookies)
        return resp(environ, start_response)

    return wrapper


def resp2flask(resp):
    """Convert an oic.utils.http_util instance to Flask."""
    if isinstance(resp, Redirect) or isinstance(resp, SeeOther):
        code = int(resp.status.split()[0])
        raise cherrypy.HTTPRedirect(resp.message, code)
    return resp.message, resp.status, resp.headers


def setup_authentication_methods(authn_config, template_env):
    """Add all authentication methods specified in the configuration."""
    routing = {}
    ac = AuthnBroker()
    for authn_method in authn_config:
        cls = make_cls_from_name(authn_method["class"])
        instance = cls(template_env=template_env, **authn_method["kwargs"])
        ac.add(authn_method["acr"], instance)
        routing[instance.url_endpoint] = VerifierMiddleware(instance)

    return ac, routing


def setup_endpoints(provider):
    """Setup the OpenID Connect Provider endpoints."""
    app_routing = {}
    endpoints = [
        AuthorizationEndpoint(
            pyoidcMiddleware(provider.authorization_endpoint)),
        TokenEndpoint(
            pyoidcMiddleware(provider.token_endpoint)),
        UserinfoEndpoint(
            pyoidcMiddleware(provider.userinfo_endpoint)),
        RegistrationEndpoint(
            pyoidcMiddleware(provider.registration_endpoint)),
        EndSessionEndpoint(
            pyoidcMiddleware(provider.endsession_endpoint))
    ]

    for ep in endpoints:
        app_routing["/{}".format(ep.etype)] = ep

    return app_routing


def _webfinger(provider, request, **kwargs):
    """Handle webfinger requests."""
    params = urlparse.parse_qs(request)
    if params["rel"][0] == OIC_ISSUER:
        wf = WebFinger()
        return Response(wf.response(params["resource"][0], provider.baseurl),
                        headers=[("Content-Type", "application/jrd+json")])
    else:
        return BadRequest("Incorrect webfinger.")


def make_static_handler(static_dir):
    def static(environ, start_response):
        path = environ['PATH_INFO']
        full_path = os.path.join(
            static_dir, os.path.normpath(path).lstrip("/"))

        if os.path.exists(full_path):
            with open(full_path, 'rb') as f:
                content = f.read()

            content_type, encoding = mimetypes.guess_type(full_path)
            headers = [('Content-Type', content_type)]
            start_response("200 OK", headers)
            return [content]
        else:
            response = NotFound(
                "File '{}' not found.".format(environ['PATH_INFO']))
            return response(environ, start_response)

    return static


def main():
    parser = argparse.ArgumentParser(description='Example OIDC Provider.')
    parser.add_argument("-p", "--port", default=80, type=int)
    parser.add_argument("-b", "--base", default="https://localhost", type=str)
    parser.add_argument("-d", "--debug", action="store_true")
    parser.add_argument("settings")
    args = parser.parse_args()

    # Load configuration
    with open(args.settings, "r") as f:
        settings = yaml.load(f)

    issuer = args.base.rstrip("/")

    template_dirs = settings["server"].get("template_dirs", "templates")
    jinja_env = Environment(loader=FileSystemLoader(template_dirs))
    authn_broker, auth_routing = setup_authentication_methods(
        settings["authn"], jinja_env)

    # Setup userinfo
    userinfo_conf = settings["userinfo"]
    cls = make_cls_from_name(userinfo_conf["class"])
    i = cls(**userinfo_conf["kwargs"])
    userinfo = UserInfo(i)

    client_db = {}
    provider = Provider(issuer, SessionDB(issuer), client_db, authn_broker,
                        userinfo, AuthzHandling(), verify_client, None)
    provider.baseurl = issuer
    provider.symkey = rndstr(16)

    # Setup keys
    path = os.path.join(os.path.dirname(__file__), "static")
    try:
        os.makedirs(path)
    except OSError, e:
        if e.errno != errno.EEXIST:
            raise e
        pass
    jwks = keyjar_init(provider, settings["provider"]["keys"])
    name = "jwks.json"
    with open(os.path.join(path, name), "w") as f:
        f.write(json.dumps(jwks))

    provider.jwks_uri.append(
        "{}/static/{}".format(provider.baseurl, name))

    # Mount the WSGI callable object (app) on the root directory
    app_routing = setup_endpoints(provider)
    app_routing["/.well-known/openid-configuration"] = pyoidcMiddleware(
        provider.providerinfo_endpoint)
    app_routing["/.well-known/webfinger"] = pyoidcMiddleware(
        partial(_webfinger, provider))
    routing = dict(auth_routing.items() + app_routing.items())
    routing["/static"] = make_static_handler(path)
    dispatcher = WSGIPathInfoDispatcher(routing)
    server = wsgiserver.CherryPyWSGIServer(('0.0.0.0', args.port), dispatcher)

    # Setup SSL
    if provider.baseurl.startswith("https://"):
        server.ssl_adapter = ssl_pyopenssl.pyOpenSSLAdapter(
            settings["server"]["cert"], settings["server"]["key"],
            settings["server"]["cert_chain"])

    # Start the CherryPy WSGI web server
    try:
        print("Server started: {}".format(issuer))
        server.start()
    except KeyboardInterrupt:
        server.stop()


if __name__ == "__main__":
    main()



"""
GFF3 parser unit tests.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import ga4gh.server.gff3 as gff3
import unittest

_testDataDir = "tests/data/datasets/dataset1/sequenceAnnotations/"


class TestGff3ParserOnTypicalFile(unittest.TestCase):
    """
    Data driven unit tests for the GFF3 parser
    """
    def setUp(self):
        testDataFile = _testDataDir + "gencodeV21Set1.gff3"
        self.gff3Parser = gff3.Gff3Parser(testDataFile)
        self.gff3Data = self.gff3Parser.parse()

    def testFileParsedHasSomeRootFeatures(self):
        self.assertIsNotNone(self.gff3Data.roots, "No root features")
        self.assertNotEqual(len(self.gff3Data.roots), 0, "No root features")

    def testSomeFeatureIsWellFormed(self):
        featId = self.gff3Data.byFeatureName.keys()[0]
        feat = self.gff3Data.byFeatureName[featId][0]
        self.assertEqual(featId, feat.featureName, "featureName mismatch")
        self.assertIsNotNone(feat.seqname, "sequence name is not populated")
        self.assertGreaterEqual(feat.end, feat.start, "end less than start")
        self.assertIn(feat.strand, u"+-", "strand is neither + nor -")
        self.assertIsNotNone(feat.source, "source is unspecified")
        self.assertIsNotNone(feat.type, "feature type is unspecified")
        self.assertIsInstance(feat.parents, set, "parents not a set")
        self.assertIsInstance(feat.children, set, "children not a set")

    def testRootFeaturesHaveNoParents(self):
        for root in self.gff3Data.roots:
            self.assertEqual(
                len(root.parents), 0, "root feature has a parent")

    def testAllFeaturesContainAllRootFeatures(self):
        for root in self.gff3Data.roots:
            feat = self.gff3Data.byFeatureName[root.featureName]
            self.assertGreaterEqual(
                len(feat), 1,
                "root feature not in list of all features")

    def testInvalidFeatureNameKeyQueryFails(self):
        badFeatureName = "987654"
        badFeat = self.gff3Data.byFeatureName[badFeatureName]
        self.assertEqual(
            len(badFeat), 0,
            "invalid feature ID returned valid object")

    def testAllChildrenFeaturesArePresentInSet(self):
        for featList in self.gff3Data.byFeatureName.values():
            for feat in featList:
                for child in feat.children:
                    childLookup = self.gff3Data.byFeatureName[
                        child.featureName]
                    self.assertGreaterEqual(
                        len(childLookup), 1,
                        "child feature not in set")


class TestGff3ParserOnDiscontinuousFeatureFile(TestGff3ParserOnTypicalFile):
    """
    Data driven parser test on file with discontinuous features.
    The tests here rely on specific data in the file being parsed.
    """
    def setUp(self):
        testDataFile = _testDataDir + "discontinuous.gff3"
        self.gff3Parser = gff3.Gff3Parser(testDataFile)
        self.gff3Data = self.gff3Parser.parse()

    def testDiscontinuousFeature(self):
        feat = self.gff3Data.byFeatureName['apidb|cds_MAL13P1.103-1']
        self.assertEqual(
            len(feat), 10,
            "not all parts of discontinuous feature parsed")


class TestGff3ParserOnSacCerFile(TestGff3ParserOnTypicalFile):
    """
    Data driven parser test on file from Saccharomyces cerevisiae S288C genome.
    """
    def setUp(self):
        testDataFile = _testDataDir + "sacCerTest.gff3"
        self.gff3Parser = gff3.Gff3Parser(testDataFile)
        self.gff3Data = self.gff3Parser.parse()


class TestGff3ParserOnSpecialCasesFile(TestGff3ParserOnTypicalFile):
    """
    Data driven parser test on a GFF3 file representing edge cases.
    """
    def setUp(self):
        testDataFile = _testDataDir + "specialCasesTest.gff3"
        self.gff3Parser = gff3.Gff3Parser(testDataFile)
        self.gff3Data = self.gff3Parser.parse()



"""
Class that builds the responses to the client
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import ga4gh.schemas.pb as pb
import ga4gh.schemas.protocol as protocol


class SearchResponseBuilder(object):
    """
    A class to allow sequential building of SearchResponse objects.
    """
    def __init__(self, responseClass, pageSize, maxBufferSize):
        """
        Allocates a new SearchResponseBuilder for the specified
        responseClass, user-requested pageSize and the system mandated
        maxBufferSize (in bytes). The maxBufferSize is an
        approximate limit on the overall length of the serialised
        response.
        """
        self._pageSize = pageSize
        self._maxBufferSize = maxBufferSize
        self._numElements = 0
        self._nextPageToken = None
        self._protoObject = responseClass()
        self._valueListName = protocol.getValueListName(responseClass)
        self._bufferSize = self._protoObject.ByteSize()

    def getPageSize(self):
        """
        Returns the page size for this SearchResponseBuilder. This is the
        user-requested maximum size for the number of elements in the
        value list.
        """
        return self._pageSize

    def getMaxBufferSize(self):
        """
        Returns the maximum internal buffer size for responses, which
        corresponds to total length (in bytes) of the serialised protobuf
        objects. This will always be less than the size of JSON output.
        """
        return self._maxBufferSize

    def getNextPageToken(self):
        """
        Returns the value of the nextPageToken for this
        SearchResponseBuilder.
        """
        return self._nextPageToken

    def setNextPageToken(self, nextPageToken):
        """
        Sets the nextPageToken to the specified value.
        """
        self._nextPageToken = nextPageToken

    def addValue(self, protocolElement):
        """
        Appends the specified protocolElement to the value list for this
        response.
        """
        self._numElements += 1
        self._bufferSize += protocolElement.ByteSize()
        attr = getattr(self._protoObject, self._valueListName)
        obj = attr.add()
        obj.CopyFrom(protocolElement)

    def isFull(self):
        """
        Returns True if the response buffer is full, and False otherwise.
        The buffer is full if either (1) the number of items in the value
        list is >= pageSize or (2) the total length of the serialised
        elements in the page is >= maxBufferSize.

        If page_size or max_response_length were not set in the request
        then they're not checked.
        """
        return (
            (self._pageSize > 0 and self._numElements >= self._pageSize) or
            (self._bufferSize >= self._maxBufferSize)
        )

    def getSerializedResponse(self):
        """
        Returns a string version of the SearchResponse that has
        been built by this SearchResponseBuilder.
        """
        self._protoObject.next_page_token = pb.string(self._nextPageToken)
        s = protocol.toJson(self._protoObject)
        return s



"""
Provides methods to initialize the server's peer to peer connections.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import requests

import ga4gh.server.datamodel as datamodel
import ga4gh.server.exceptions as exceptions


def getInitialPeerList(filePath, logger=None):
    """
    Attempts to get a list of peers from a file specified in configuration.

    This file has one URL per line and can contain newlines and comments.

        # Main ga4gh node
        http://1kgenomes.ga4gh.org
        # Local intranet peer
        https://192.168.1.1

    The server will attempt to add URLs in this file to its registry at
    startup and will log a warning if the file isn't found.
    """
    ret = []
    with open(filePath) as textFile:
        ret = textFile.readlines()
    if len(ret) == 0:
        if logger:
            logger.warn("Couldn't load the initial "
                        "peer list. Try adding a "
                        "file named 'initial_peers.txt' "
                        "to {}".format(os.getcwd()))
    # Remove lines that start with a hash or are empty.
    return filter(lambda x: x != "" and not x.find("#") != -1, ret)


def insertInitialPeer(dataRepository, url, logger=None):
    """
    Takes the datarepository, a url, and an optional logger and attempts
    to add the peer into the repository.
    """
    insertPeer = dataRepository.insertPeer
    try:
        peer = datamodel.peers.Peer(url)
        insertPeer(peer)
    except exceptions.RepoManagerException as exc:
        if logger:
            logger.debug(
                "Peer already in registry {} {}".format(peer.getUrl(), exc))
    except exceptions.BadUrlException as exc:
        if logger:
            logger.debug("A URL in the initial "
                         "peer list {} was malformed. {}".format(url), exc)


def initialize(filePath, dataRepository, logger=None):
    for initialPeer in getInitialPeerList(filePath):
        try:
            headers = {'content-type': 'application/json'}
            data = {"peer": {"url": initialPeer}}
            url = "{}/announce".format(initialPeer.rstrip("/"))
            requests.post(url, headers=headers, json=data)
        except Exception as e:
            if logger:
                logger.info("Couldn't announce to initial peer {}".format(
                    (e, url)))
        insertInitialPeer(dataRepository, initialPeer, logger)



"""
Glue to enable script access to ga4gh packages
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
# the following two lines are the same ones as in dev_glue.py
# they enable python to find the ga4gh.server package
import ga4gh
ga4gh.__path__.insert(0, 'ga4gh')


def ga4ghImportGlue():
    """
    Call this method before importing a ga4gh module in the scripts dir.
    Otherwise, you will be using the installed package instead of
    the development package.
    Assumes a certain directory structure.
    """
    path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    sys.path.append(path)



"""
Builds the test data registry DB.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import glob
import os.path

import ga4gh.common.utils as utils


def run(*args):
    cmd = "python repo_dev.py {}".format(" ".join(args))
    print("running:", cmd)
    utils.runCommand(cmd)


def buildTestData(
        dataDirectory='tests/data', relativePaths=False, force=False):
    prefix = dataDirectory
    repoFile = os.path.join(prefix, "registry.db")
    if os.path.exists(repoFile):
        if force:
            print("deleting repo at '{}'".format(repoFile))
            os.unlink(repoFile)
        else:
            print("'{}' already exists".format(repoFile))
            return
    print("building repo at '{}'".format(repoFile))
    sequenceOntologyName = "so-xp-simple"
    useRelativePath = '-r' if relativePaths else ''
    run("init", "-f", repoFile)

    run("add-peer", repoFile, useRelativePath, "http://example.ga4gh.org")

    pattern = os.path.join(prefix, "referenceSets", "*.fa.gz")
    for dataFile in glob.glob(pattern):
        run("add-referenceset", repoFile, useRelativePath, dataFile,
            "--species ", '\'{"termId": "NCBI:9606", '
                          '"term": "Homo sapiens"}\'')

    pattern = os.path.join(prefix, "ontologies", "*.obo")
    for dataFile in glob.glob(pattern):
        run("add-ontology", repoFile, useRelativePath, dataFile)

    datasetName = "dataset1"
    run("add-dataset", repoFile, datasetName)

    pattern = os.path.join(prefix, "datasets/dataset1/reads", "*.bam")
    for dataFile in glob.glob(pattern):
        run("add-readgroupset", repoFile, datasetName, useRelativePath,
            dataFile)

    pattern = os.path.join(prefix, "datasets/dataset1/variants", "*")
    for j, dataFile in enumerate(glob.glob(pattern)):
        name = "vs_{}".format(j)
        run(
            "add-variantset", repoFile, datasetName, useRelativePath,
            dataFile, "-R NCBI37", "-n ", name, "-aO", sequenceOntologyName)

    pattern = os.path.join(
        prefix, "datasets/dataset1/sequenceAnnotations", "*.db")
    for j, dataFile in enumerate(glob.glob(pattern)):
        run(
            "add-featureset", repoFile, datasetName, useRelativePath,
            dataFile, "-R NCBI37", "-O", sequenceOntologyName,
            "-C ga4gh.datamodel.sequence_annotations.Gff3DbFeatureSet")

    pattern = os.path.join(
        prefix, "datasets/dataset1/continuous", "*.bw")
    for dataFile in glob.glob(pattern):
        run("add-continuousset", repoFile, datasetName, useRelativePath,
            dataFile, "-R NCBI37",
            "-C ga4gh.datamodel.continuous.FileContinuousSet")

    pattern = os.path.join(prefix, "datasets/dataset1/phenotypes", "*")
    for dataFile in glob.glob(pattern):
        # coordinate featureset name and g2p name
        name = dataFile.split("/")[-1]
        run(
            "add-phenotypeassociationset", repoFile,
            datasetName, dataFile, "-n {}".format(name))
        run(
            "add-featureset", repoFile, datasetName, useRelativePath,
            dataFile, "-R NCBI37",  "-O", sequenceOntologyName,
            "-C ga4gh.datamodel.genotype_phenotype_featureset."
            "PhenotypeAssociationFeatureSet")

    pattern = os.path.join(
        prefix, "datasets/dataset1/rnaQuant", "*.db")
    for j, dataFile in enumerate(glob.glob(pattern)):
        name = "rnaseq_{}".format(j)
        run(
            "add-rnaquantificationset", repoFile, datasetName, dataFile,
            "-R NCBI37", "-n ", name)


def parseArgs():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d", "--data-directory", default='tests/data', action="store_true",
        help="root directory for test data")
    parser.add_argument(
        "-r", "--relativePaths", default=False, action="store_true",
        help="store relative paths in database")
    parser.add_argument(
        "-f", "--force", default=False, action="store_true",
        help="delete previous database and build a new one")
    args = parser.parse_args()
    return args


@utils.Timed()
def main():
    args = parseArgs()
    buildTestData(args.data_directory, args.relativePaths, args.force)


if __name__ == "__main__":
    main()



"""
Unit tests for the frontend code.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import shutil
import unittest
import logging

import ga4gh.server.frontend as frontend
import ga4gh.server.exceptions as exceptions
import ga4gh.server.auth as auth

import ga4gh.schemas.protocol as protocol


class TestAuth0(unittest.TestCase):
    """
    Tests the basic routing and HTTP handling for the Flask app.
    """
    exampleUrl = 'www.example.com'

    @classmethod
    def setUpClass(cls):
        config = {
            "DATA_SOURCE": "simulated://",
            "SIMULATED_BACKEND_RANDOM_SEED": 1111,
            "SIMULATED_BACKEND_NUM_CALLS": 1,
            "SIMULATED_BACKEND_VARIANT_DENSITY": 1.0,
            "SIMULATED_BACKEND_NUM_VARIANT_SETS": 1,
            "CACHE_DIRECTORY": "/tmp/ga4gh-test"
        }
        frontend.reset()
        frontend.configure(
            baseConfig="TestAuth0Config", extraConfig=config)
        cls.app = frontend.app
        # silence usually unhelpful CORS log
        logging.getLogger('ga4gh.frontend.cors').setLevel(logging.CRITICAL)
        cls.backend = frontend.app.backend
        cls.client = frontend.app.test_client()

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree('/tmp/ga4gh-test', True)
        frontend.reset()

    def sendPostRequest(self, path, request, extraHeaders=None):
        """
        Sends the specified GA request object and returns the response.
        """
        headers = {
            'Content-type': 'application/json',
            'Origin': self.exampleUrl,
        }
        if extraHeaders:
            headers.update(extraHeaders)
        return self.client.post(
            path, headers=headers, data=protocol.toJson(request))

    def sendGetRequest(self, path):
        """
        Sends a get request to the specified URL and returns the response.
        """
        headers = {
            'Origin': self.exampleUrl,
        }
        return self.client.get(path, headers=headers)

    def testCallback(self):
        response = self.sendGetRequest("callback")
        self.assertEqual(
            response.status_code,
            401, "Ensure that when the callback is hit without a code"
                 "it will return 401 but got {}".format(response.status_code))
        response = self.sendGetRequest("callback?code=abc")
        self.assertEqual(
            response.status_code,
            401, "Ensure that when the callback is hit without a code"
                 "it will return 401 but got {}".format(response.status_code))

    def testLogin(self):
        response = self.sendGetRequest("login")
        self.assertEqual(
            response.status_code,
            200, "Ensure that when Auth0 is turned on the login page"
                 "returns 200 {}".format(response.status_code))

    def testBadBearer(self):
        """
        Tests to see if a malformed bearer token fails in expected ways
        """
        response = self.client.get('/')
        self.assertEqual(response.status_code, 401)
        protectedPath = "datasets/search"
        request = protocol.SearchDatasetsRequest()
        headers = {"Authorization": ""}
        response = self.sendPostRequest(protectedPath, request, headers)
        self.assertEquals(response.status_code, 401, "No bearer should fail"
                                                     "with 401")
        headers = {"Authorization": "Bearer"}
        response = self.sendPostRequest(protectedPath, request, headers)
        self.assertEquals(response.status_code, 401, "")

    def testProtected(self):
        protectedPath = "datasets/search"
        request = protocol.SearchDatasetsRequest()
        response = self.sendPostRequest(protectedPath, request)
        self.assertEquals(
            response.status_code,
            401, "If Auth0 is enabled this endpoint "
                 "is not accessible without auth.")

    def testDecodeExceptions(self):
        """
        Unit tests the header parsing functions.
        """
        with self.assertRaises(exceptions.NotAuthorizedException):
            auth._has_header('')
        with self.assertRaises(exceptions.NotAuthorizedException):
            auth._has_bearer('empty')
        with self.assertRaises(exceptions.NotAuthorizedException):
            auth._has_token('Bearer')
        with self.assertRaises(exceptions.NotAuthorizedException):
            auth._has_token('Bearer123')
        with self.assertRaises(exceptions.NotAuthorizedException):
            auth._well_formed('Bearer 123 456')
        with self.assertRaises(exceptions.NotAuthorizedException):
            client_id = self.app.config.get("AUTH0_CLIENT_ID")
            client_secret = self.app.config.get("AUTH0_CLIENT_SECRET")
            token, payload = auth._decode_header(
                'Bearer 123', client_id, client_secret)

    def testRenderLogin(self):
        """
        Tests that the login template renders without failing
        """
        return auth.render_login(
            app=self.app,
            scopes=self.app.config.get('AUTH0_SCOPES'),
            redirect_uri=self.app.config.get('AUTH0_CALLBACK_URL'),
            domain=self.app.config.get('AUTH0_HOST'),
            client_id=self.app.config.get('AUTH0_CLIENT_ID'))

    def testAuthorizeEmail(self):
        """
        Tests that the email is set to authorized in the cache.
        """
        email = "test@test.it"
        auth.authorize_email(email, self.app.cache)
        entry = self.app.cache.get(email)
        self.assertTrue(entry['authorized'])
        self.assertTrue(auth.is_authorized(self.app.cache, email))

    def testIsActive(self):
        """
        Tests that is active throws an exception when the token is not found.
        """
        token = "123"
        email = "test@test.com"
        with self.assertRaises(exceptions.NotAuthenticatedException):
            auth.is_active(self.app.cache, token)
        self.app.cache.set("123", {"email": email})
        # Even though the token is set, in the cache, it
        self.assertEquals(
            email, auth.is_active(self.app.cache, token)['email'])

    def testCallbackMaker(self):
        """
        Tests that the callback maker returns a function and runs without
         failure.
        """
        callback = auth.callback_maker(
            cache=self.app.cache,
            domain=self.app.config.get('AUTH0_HOST'),
            client_id=self.app.config.get('AUTH0_CLIENT_ID'),
            client_secret=self.app.config.get('AUTH0_CLIENT_SECRET'),
            redirect_uri=self.app.config.get('AUTH0_CALLBACK_URL'))
        self.assertTrue(callable(callback))



"""
Tests for the backend objects. We instantiate local copies of
the backends and invoke the entry points for the protocol methods.
We do not set up any server processes or communicate over sockets.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest

import ga4gh.server.exceptions as exceptions
import ga4gh.server.backend as backend
import ga4gh.server.paging as paging
import ga4gh.server.datarepo as datarepo
import ga4gh.server.datamodel.datasets as datasets
import ga4gh.server.datamodel.references as references

import tests.paths as paths


class TestAbstractBackend(unittest.TestCase):
    """
    Tests for shared functionality between backends.
    """
    def setUp(self):
        self._backend = backend.Backend(datarepo.AbstractDataRepository())
        self._dataRepo = self._backend.getDataRepository()

    def testAddOneDataset(self):
        datasetName = "ds"
        dataset = datasets.Dataset(datasetName)
        self.assertEqual(self._dataRepo.getNumDatasets(), 0)
        self.assertEqual(self._dataRepo.getDatasets(), [])
        self._dataRepo.addDataset(dataset)
        self.assertEqual(self._dataRepo.getNumDatasets(), 1)
        self.assertEqual(self._dataRepo.getDatasets(), [dataset])
        self.assertEqual(self._dataRepo.getDatasetByIndex(0), dataset)
        self.assertEqual(self._dataRepo.getDatasetByName(datasetName), dataset)
        self.assertEqual(self._dataRepo.getDataset(dataset.getId()), dataset)

    def testAddMultipleDatasets(self):
        firstDatasetName = "ds1"
        firstDataset = datasets.Dataset(firstDatasetName)
        secondDatasetName = "ds2"
        secondDataset = datasets.Dataset(secondDatasetName)
        self.assertEqual(self._dataRepo.getNumDatasets(), 0)
        self.assertEqual(self._dataRepo.getDatasets(), [])
        self._dataRepo.addDataset(firstDataset)
        self._dataRepo.addDataset(secondDataset)
        self.assertEqual(self._dataRepo.getNumDatasets(), 2)
        self.assertEqual(self._dataRepo.getDatasets(),
                         [firstDataset, secondDataset])
        self.assertEqual(self._dataRepo.getDatasetByIndex(0), firstDataset)
        self.assertEqual(self._dataRepo.getDatasetByIndex(1), secondDataset)
        self.assertEqual(self._dataRepo.getDatasetByName(firstDatasetName),
                         firstDataset)
        self.assertEqual(self._dataRepo.getDatasetByName(secondDatasetName),
                         secondDataset)
        self.assertEqual(self._dataRepo.getDataset(firstDataset.getId()),
                         firstDataset)
        self.assertEqual(self._dataRepo.getDataset(secondDataset.getId()),
                         secondDataset)

    def testAddOneReferenceSet(self):
        referenceSetLocalId = "id"
        referenceSet = references.AbstractReferenceSet(referenceSetLocalId)
        self.assertEqual(self._dataRepo.getNumReferenceSets(), 0)
        self.assertEqual(self._dataRepo.getReferenceSets(), [])
        self._dataRepo.addReferenceSet(referenceSet)
        self.assertEqual(self._dataRepo.getNumReferenceSets(), 1)
        self.assertEqual(self._dataRepo.getReferenceSets(), [referenceSet])
        self.assertEqual(
            self._dataRepo.getReferenceSetByIndex(0), referenceSet)
        self.assertEqual(
            self._dataRepo.getReferenceSetByName(referenceSet.getLocalId()),
            referenceSet)
        self.assertEqual(self._dataRepo.getReferenceSet(referenceSet.getId()),
                         referenceSet)

    def testAddMultipleReferenceSet(self):
        firstRSLocalId = "id1"
        firstRS = references.AbstractReferenceSet(firstRSLocalId)
        secondRSLocalId = "id2"
        secondRS = references.AbstractReferenceSet(secondRSLocalId)
        self.assertEqual(self._dataRepo.getNumReferenceSets(), 0)
        self.assertEqual(self._dataRepo.getReferenceSets(), [])
        self._dataRepo.addReferenceSet(firstRS)
        self._dataRepo.addReferenceSet(secondRS)
        self.assertEqual(self._dataRepo.getNumReferenceSets(), 2)
        self.assertEqual(self._dataRepo.getReferenceSets(),
                         [firstRS, secondRS])
        self.assertEqual(self._dataRepo.getReferenceSetByIndex(0),
                         firstRS)
        self.assertEqual(self._dataRepo.getReferenceSetByIndex(1),
                         secondRS)
        self.assertEqual(
            self._dataRepo.getReferenceSetByName(firstRS.getLocalId()),
            firstRS)
        self.assertEqual(
            self._dataRepo.getReferenceSetByName(secondRS.getLocalId()),
            secondRS)
        self.assertEqual(self._dataRepo.getReferenceSet(firstRS.getId()),
                         firstRS)
        self.assertEqual(self._dataRepo.getReferenceSet(secondRS.getId()),
                         secondRS)

    def testGetDatasetBadId(self):
        for badId in ["", None, "NO SUCH ID"]:
            self.assertRaises(
                exceptions.DatasetNotFoundException,
                self._dataRepo.getDataset, badId)

    def testGetReferenceSetBadId(self):
        for badId in ["", None, "NO SUCH ID"]:
            self.assertRaises(
                exceptions.ReferenceSetNotFoundException,
                self._dataRepo.getReferenceSet, badId)

    def testGetDatasetBadName(self):
        for badName in ["", None, "NO SUCH NAME"]:
            self.assertRaises(
                exceptions.DatasetNameNotFoundException,
                self._dataRepo.getDatasetByName, badName)

    def testGetReferenceSetBadName(self):
        for badName in ["", None, "NO SUCH NAME"]:
            self.assertRaises(
                exceptions.ReferenceSetNameNotFoundException,
                self._dataRepo.getReferenceSetByName, badName)

    def testGetDatasetByIndexBadIndex(self):
        self.assertRaises(IndexError, self._dataRepo.getDatasetByIndex, 0)
        self.assertRaises(TypeError, self._dataRepo.getDatasetByIndex, None)
        self.assertRaises(TypeError, self._dataRepo.getDatasetByIndex, "")
        datasetName = "ds"
        dataset = datasets.Dataset(datasetName)
        self._dataRepo.addDataset(dataset)
        self.assertRaises(IndexError, self._dataRepo.getDatasetByIndex, 1)

    def testGetReferenceSetByIndexBadIndex(self):
        self.assertRaises(IndexError, self._dataRepo.getReferenceSetByIndex, 0)
        self.assertRaises(TypeError,
                          self._dataRepo.getReferenceSetByIndex, None)
        self.assertRaises(TypeError, self._dataRepo.getReferenceSetByIndex, "")
        referenceSetName = "id"
        referenceSet = references.AbstractReferenceSet(referenceSetName)
        self._dataRepo.addReferenceSet(referenceSet)
        self.assertRaises(IndexError, self._dataRepo.getReferenceSetByIndex, 1)


class TestSqlRepoTestData(unittest.TestCase):
    """
    Tests proper initialization of the SQL repo based on known
    files in the tests/data directory.
    """
    def setUp(self):
        self._dataRepo = datarepo.SqlDataRepository(paths.testDataRepo)
        self._dataRepo.open(datarepo.MODE_READ)

    def testDatasets(self):
        self.assertEqual(self._dataRepo.getNumDatasets(), 1)
        dataset = self._dataRepo.getDatasetByIndex(0)
        self.assertEqual(dataset.getLocalId(), "dataset1")
        self.assertEqual(self._dataRepo.getDatasetByName("dataset1"), dataset)

    def testReferenceSets(self):
        self.assertEqual(self._dataRepo.getNumReferenceSets(), 4)
        referenceSets = enumerate(self._dataRepo.getReferenceSets())
        referenceSetsByName = sorted(
            referenceSets, key=lambda x: x[1].getLocalId())
        expected_names = sorted(
            ["Default", "NCBI37", "example_1", "example_2"])
        for name, (index, rs) in zip(expected_names, referenceSetsByName):
            self.assertEqual(rs.getLocalId(), name)
            self.assertEqual(self._dataRepo.getReferenceSetByIndex(index), rs)
            self.assertEqual(self._dataRepo.getReferenceSet(rs.getId()), rs)
            self.assertEqual(self._dataRepo.getReferenceSetByName(name), rs)


class TestTopLevelObjectGenerator(unittest.TestCase):
    """
    Tests the generator used for top level objects
    """
    def setUp(self):
        class FakeRequest(object):
            pass

        class FakeTopLevelObject(object):
            def toProtocolElement(self):
                return self

        self.request = FakeRequest()
        self.request.page_token = ""
        self.num_objects = 3
        self.objects = [FakeTopLevelObject() for j in range(self.num_objects)]
        self.backend = backend.Backend(datarepo.AbstractDataRepository())

    def getObjectByIndex(self, index):
        return self.objects[index]

    def testPageToken(self):
        self.request.page_token = "1"
        self._assertNumItems(2)

    def testPageTokenNone(self):
        self._assertNumItems(3)

    def _assertNumItems(self, numItems):
        iterator = self.backend._topLevelObjectGenerator(
            self.request, self.num_objects, self.getObjectByIndex)
        items = list(iterator)
        self.assertEqual(len(items), numItems)


class TestPrivateBackendMethods(unittest.TestCase):
    """
    keep tests of private backend methods here and not in one of the
    subclasses of TestAbstractBackend, otherwise the tests will needlessly
    be run more than once

    (they could be put in TestAbstractBackend, but I think it's a clearer
    separation to put them in their own test class)
    """
    def testParsePageToken(self):
        goodPageToken = "12:34:567:8:9000"
        parsedToken = paging._parsePageToken(goodPageToken, 5)
        self.assertEqual(parsedToken[2], 567)

    def testParseIntegerArgument(self):
        good = {"one": "1", "minusone": "-1"}
        expected = {"one": 1, "minusone": -1}
        bad = {"string": "A", "float": "0.98"}
        self.assertEqual(paging._parseIntegerArgument(
            {}, "missing", 0), 0)
        for key in good:
            self.assertEqual(
                paging._parseIntegerArgument(
                    good, key, 0), expected[key])
        for key in bad:
            with self.assertRaises(exceptions.BadRequestIntegerException):
                paging._parseIntegerArgument(bad, key, 0)



import importlib
from oic.utils.authn.user import UserAuthnMethod

__author__ = 'regu0004'


class AuthnModule(UserAuthnMethod):

    # override in subclass specifying suitable url endpoint to POST user input
    url_endpoint = "/verify"
    FAILED_AUTHN = (None, True)

    def __call__(self, *args, **kwargs):
        """
        Display user interaction.
        :return: instance of oic.utils.http_util.Response
        """
        raise NotImplementedError()

    def verify(self, *args, **kwargs):
        """
        Callback to verify user input
        :return: username of the authenticated user
        """
        raise NotImplementedError()


def make_cls_from_name(name):
    module_name, cls_name = name.rsplit(".", 1)
    cls = getattr(importlib.import_module(module_name), cls_name)
    return cls



"""
Simple shim for running the repo program during development.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import dev_glue  # NOQA
import ga4gh.server.cli.repomanager as cli_repomanager

if __name__ == "__main__":
    cli_repomanager.repo_main()



"""
Module responsible for translating variant data into GA4GH native
objects.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import datetime
import glob
import hashlib
import json
import os
import random
import re

import pysam

import ga4gh.server.exceptions as exceptions
import ga4gh.server.datamodel as datamodel

import ga4gh.schemas.pb as pb
import ga4gh.schemas.protocol as protocol


ANNOTATIONS_VEP_V82 = "VEP_v82"
ANNOTATIONS_VEP_V77 = "VEP_v77"
ANNOTATIONS_SNPEFF = "SNPEff"


# Utility functions for module

def isUnspecified(str):
    """
    Checks whether a string is None or an
    empty string. Returns a boolean.
    """
    return str == "" or str is None


_nothing = object()


def isEmptyIter(it):
    """Return True iff the iterator is empty or exhausted"""
    return next(it, _nothing) is _nothing


class CallSet(datamodel.DatamodelObject):
    """
    Class representing a CallSet. A CallSet basically represents the
    metadata associated with a single VCF sample column.
    """
    compoundIdClass = datamodel.CallSetCompoundId

    def __init__(self, parentContainer, localId):
        super(CallSet, self).__init__(parentContainer, localId)
        self._info = {}
        self._biosampleId = None

    def populateFromRow(self, callSetRecord):
        """
        Populates this CallSet from the specified DB row.
        """
        self._biosampleId = callSetRecord.biosampleid
        self.setAttributesJson(callSetRecord.attributes)

    def toProtocolElement(self):
        """
        Returns the representation of this CallSet as the corresponding
        ProtocolElement.
        """
        variantSet = self.getParentContainer()
        gaCallSet = protocol.CallSet(
            biosample_id=self.getBiosampleId())
        if variantSet.getCreationTime():
            gaCallSet.created = variantSet.getCreationTime()
        if variantSet.getUpdatedTime():
            gaCallSet.updated = variantSet.getUpdatedTime()
        gaCallSet.id = self.getId()
        gaCallSet.name = self.getLocalId()
        gaCallSet.variant_set_ids.append(variantSet.getId())
        self.serializeAttributes(gaCallSet)
        return gaCallSet

    def getBiosampleId(self):
        """
        Returns the biosampleId for this CallSet.
        """
        return self._biosampleId

    def setBiosampleId(self, biosampleId):
        """
        Set the biosampleId for the current sample.
        """
        self._biosampleId = biosampleId

    def getSampleName(self):
        """
        Returns the sample name for this CallSet.
        """
        return self.getLocalId()

    def getInfo(self):
        """
        Returns info map for this CallSet
        """
        return self._info


class AbstractVariantSet(datamodel.DatamodelObject):
    """
    An abstract base class of a variant set
    """
    compoundIdClass = datamodel.VariantSetCompoundId

    def __init__(self, parentContainer, localId):
        super(AbstractVariantSet, self).__init__(parentContainer, localId)
        self._callSetIdMap = {}
        self._callSetNameMap = {}
        self._callSetIds = []
        self._callSetIdToIndex = {}
        self._creationTime = None
        self._updatedTime = None
        self._referenceSet = None
        self._metadata = []
        self._variantAnnotationSetIds = []
        self._variantAnnotationSetIdMap = {}

    def addVariantAnnotationSet(self, variantAnnotationSet):
        """
        Adds the specified variantAnnotationSet to this dataset.
        """
        id_ = variantAnnotationSet.getId()
        self._variantAnnotationSetIdMap[id_] = variantAnnotationSet
        self._variantAnnotationSetIds.append(id_)

    def getVariantAnnotationSets(self):
        """
        Returns the list of VariantAnnotationSets in this dataset
        """
        return [
            self._variantAnnotationSetIdMap[id_] for id_ in
            self._variantAnnotationSetIds]

    def getVariantAnnotationSet(self, id_):
        """
        Returns the AnnotationSet in this dataset with the specified 'id'
        """
        if id_ not in self._variantAnnotationSetIdMap:
            raise exceptions.AnnotationSetNotFoundException(id_)
        return self._variantAnnotationSetIdMap[id_]

    def getNumVariantAnnotationSets(self):
        """
        Returns the number of variant annotation sets in this dataset.
        """
        return len(self._variantAnnotationSetIds)

    def getVariantAnnotationSetByIndex(self, index):
        """
        Returns the variant annotation set at the specified index in this
        dataset.
        """
        return self._variantAnnotationSetIdMap[
            self._variantAnnotationSetIds[index]]

    def setReferenceSet(self, referenceSet):
        """
        Sets the ReferenceSet for this VariantSet to the specified value.
        """
        self._referenceSet = referenceSet

    def getReferenceSet(self):
        """
        Returns the reference set associated with this VariantSet.
        """
        return self._referenceSet

    def getCreationTime(self):
        """
        Returns the creation time for this variant set.
        """
        return self._creationTime

    def getUpdatedTime(self):
        """
        Returns the time this variant set was last updated.
        """
        return self._updatedTime

    def addCallSet(self, callSet):
        """
        Adds the specfied CallSet to this VariantSet.
        """
        callSetId = callSet.getId()
        self._callSetIdMap[callSetId] = callSet
        self._callSetNameMap[callSet.getLocalId()] = callSet
        self._callSetIds.append(callSetId)
        self._callSetIdToIndex[callSet.getId()] = len(self._callSetIds) - 1

    def addCallSetFromName(self, sampleName):
        """
        Adds a CallSet for the specified sample name.
        """
        callSet = CallSet(self, sampleName)
        self.addCallSet(callSet)

    def getCallSets(self):
        """
        Returns the list of CallSets in this VariantSet.
        """
        return [self._callSetIdMap[id_] for id_ in self._callSetIds]

    def getNumCallSets(self):
        """
        Returns the number of CallSets in this variant set.
        """
        return len(self._callSetIds)

    def getCallSetByName(self, name):
        """
        Returns a CallSet with the specified name, or raises a
        CallSetNameNotFoundException if it does not exist.
        """
        if name not in self._callSetNameMap:
            raise exceptions.CallSetNameNotFoundException(name)
        return self._callSetNameMap[name]

    def getCallSetByIndex(self, index):
        """
        Returns the CallSet at the specfied index in this VariantSet.
        """
        return self._callSetIdMap[self._callSetIds[index]]

    def getCallSet(self, id_):
        """
        Returns a CallSet with the specified id, or raises a
        CallSetNotFoundException if it does not exist.
        """
        if id_ not in self._callSetIdMap:
            raise exceptions.CallSetNotFoundException(id_)
        return self._callSetIdMap[id_]

    def getMetadata(self):
        """
        Returns Metdata associated with this VariantSet
        """
        return self._metadata

    def toProtocolElement(self):
        """
        Converts this VariantSet into its GA4GH protocol equivalent.
        """
        protocolElement = protocol.VariantSet()
        protocolElement.id = self.getId()
        protocolElement.dataset_id = self.getParentContainer().getId()
        protocolElement.reference_set_id = self._referenceSet.getId()
        protocolElement.metadata.extend(self.getMetadata())
        protocolElement.dataset_id = self.getParentContainer().getId()
        protocolElement.reference_set_id = self._referenceSet.getId()
        protocolElement.name = self.getLocalId()
        self.serializeAttributes(protocolElement)
        return protocolElement

    def getNumVariants(self):
        """
        Returns the number of variants contained in this VariantSet.
        """
        raise NotImplementedError()

    def _createGaVariant(self):
        """
        Convenience method to set the common fields in a GA Variant
        object from this variant set.
        """
        ret = protocol.Variant()
        if self._creationTime:
            ret.created = self._creationTime
        if self._updatedTime:
            ret.updated = self._updatedTime
        ret.variant_set_id = self.getId()
        return ret

    def getVariantId(self, gaVariant):
        """
        Returns an ID string suitable for the specified GA Variant
        object in this variant set.
        """
        md5 = self.hashVariant(gaVariant)
        compoundId = datamodel.VariantCompoundId(
            self.getCompoundId(), gaVariant.reference_name,
            str(gaVariant.start), md5)
        return str(compoundId)

    def getCallSetId(self, sampleName):
        """
        Returns the callSetId for the specified sampleName in this
        VariantSet.
        """
        compoundId = datamodel.CallSetCompoundId(
            self.getCompoundId(), sampleName)
        return str(compoundId)

    @classmethod
    def hashVariant(cls, gaVariant):
        """
        Produces an MD5 hash of the ga variant object to distinguish
        it from other variants at the same genomic coordinate.
        """
        hash_str = gaVariant.reference_bases + \
            str(tuple(gaVariant.alternate_bases))
        return hashlib.md5(hash_str).hexdigest()


class SimulatedVariantSet(AbstractVariantSet):
    """
    A variant set that doesn't derive from a data store.
    Used mostly for testing.
    """
    def __init__(
            self, parentContainer, referenceSet, localId, randomSeed=1,
            numCalls=1, variantDensity=1):
        super(SimulatedVariantSet, self).__init__(parentContainer, localId)
        self._referenceSet = referenceSet
        self._randomSeed = randomSeed
        self._numCalls = numCalls
        for i in range(numCalls):
            callSetName = "simCallSet_{}".format(i)
            self.addCallSetFromName(callSetName)
            callSet = self.getCallSetByName(callSetName)
            # build up infos of increasing size
            for j in range(i):
                callSet._info["key_{}".format(j)] = "value_{}".format(j)
        self._variantDensity = variantDensity
        self._metadata = self._createMetaData()
        now = protocol.convertDatetime(datetime.datetime.now())
        self._creationTime = now
        self._updatedTime = now

    def _createMetaData(self):
        metadata_1 = protocol.VariantSetMetadata()
        metadata_1.key = "version"
        metadata_1.value = "VCFv4.1"
        metadata_1.type = "String"
        metadata_1.number = "1"
        metadata_1.description = ""
        metadata_1.id = str(datamodel.VariantSetMetadataCompoundId(
            self.getCompoundId(), 'metadata:' + metadata_1.key))

        metadata_2 = protocol.VariantSetMetadata()
        metadata_2.key = "INFO.FIELD1"
        metadata_2.value = ""
        metadata_2.type = "String"
        metadata_2.number = ""
        metadata_2.description = "FIELD1 description"
        metadata_2.id = str(datamodel.VariantSetMetadataCompoundId(
            self.getCompoundId(), 'metadata:' + metadata_2.key))

        metadata_3 = protocol.VariantSetMetadata()
        metadata_3.key = "INFO.FIELD2"
        metadata_3.value = ""
        metadata_3.type = "Integer"
        metadata_3.number = "1"
        metadata_3.description = "FIELD2 description"
        metadata_3.id = str(datamodel.VariantSetMetadataCompoundId(
            self.getCompoundId(), 'metadata:' + metadata_3.key))

        return [metadata_1, metadata_2, metadata_3]

    def getNumVariants(self):
        return 0

    def getVariant(self, compoundId):
        randomNumberGenerator = random.Random()
        start = int(compoundId.start)
        randomNumberGenerator.seed(self._randomSeed + start)
        variant = self.generateVariant(
            compoundId.reference_name, start, randomNumberGenerator)
        return variant

    def getVariants(self, referenceName, startPosition, endPosition,
                    callSetIds=None):
        randomNumberGenerator = random.Random()
        randomNumberGenerator.seed(self._randomSeed)
        i = startPosition
        while i < endPosition:
            if randomNumberGenerator.random() < self._variantDensity:
                randomNumberGenerator.seed(self._randomSeed + i)
                yield self.generateVariant(
                    referenceName, i, randomNumberGenerator)
            i += 1

    def generateVariant(self, referenceName, position, randomNumberGenerator):
        """
        Generate a random variant for the specified position using the
        specified random number generator. This generator should be seeded
        with a value that is unique to this position so that the same variant
        will always be produced regardless of the order it is generated in.
        """
        variant = self._createGaVariant()
        variant.reference_name = referenceName
        variant.start = position
        variant.end = position + 1  # SNPs only for now
        bases = ["A", "C", "G", "T"]
        ref = randomNumberGenerator.choice(bases)
        variant.reference_bases = ref
        alt = randomNumberGenerator.choice(
            [base for base in bases if base != ref])
        variant.alternate_bases.append(alt)
        randChoice = randomNumberGenerator.randint(0, 2)
        if randChoice == 0:
            variant.filters_applied = False
        elif randChoice == 1:
            variant.filters_applied = True
            variant.filters_passed = True
        else:
            variant.filters_applied = True
            variant.filters_passed = False
            variant.filters_failed.append('q10')
        for callSet in self.getCallSets():
            call = variant.calls.add()
            call.call_set_id = callSet.getId()
            # for now, the genotype is either [0,1], [1,1] or [1,0] with equal
            # probability; probably will want to do something more
            # sophisticated later.
            randomChoice = randomNumberGenerator.choice(
                [[0, 1], [1, 0], [1, 1]])
            call.genotype.extend(randomChoice)
            # TODO What is a reasonable model for generating these likelihoods?
            # Are these log-scaled? Spec does not say.
            call.genotype_likelihood.extend([-100, -100, -100])
        variant.id = self.getVariantId(variant)
        return variant


class HtslibVariantSet(datamodel.PysamDatamodelMixin, AbstractVariantSet):
    """
    Class representing a single variant set backed by a directory of indexed
    VCF or BCF files.
    """
    def __init__(self, parentContainer, localId):
        super(HtslibVariantSet, self).__init__(parentContainer, localId)
        self._chromFileMap = {}
        self._metadata = None

    def isAnnotated(self):
        """
        Returns True if there is a VariantAnnotationSet associated with this
        VariantSet.
        """
        return len(self._variantAnnotationSetIdMap) > 0

    def getReferenceToDataUrlIndexMap(self):
        """
        Returns the map of Reference names to the (dataUrl, indexFile) pairs.
        """
        return self._chromFileMap

    def getDataUrlIndexPairs(self):
        """
        Returns the set of (dataUrl, indexFile) pairs.
        """
        return set(self._chromFileMap.values())

    def populateFromRow(self, variantSetRecord):
        """
        Populates this VariantSet from the specified DB row.
        """
        self._created = variantSetRecord.created
        self._updated = variantSetRecord.updated
        self.setAttributesJson(variantSetRecord.attributes)
        self._chromFileMap = {}
        # We can't load directly as we want tuples to be stored
        # rather than lists.
        for key, value in json.loads(variantSetRecord.dataurlindexmap).items():
            self._chromFileMap[key] = tuple(value)
        self._metadata = []
        for jsonDict in json.loads(variantSetRecord.metadata):
            metadata = protocol.fromJson(json.dumps(jsonDict),
                                         protocol.VariantSetMetadata)
            self._metadata.append(metadata)

    def populateFromFile(self, dataUrls, indexFiles):
        """
        Populates this variant set using the specified lists of data
        files and indexes. These must be in the same order, such that
        the jth index file corresponds to the jth data file.
        """
        assert len(dataUrls) == len(indexFiles)
        for dataUrl, indexFile in zip(dataUrls, indexFiles):
            varFile = pysam.VariantFile(dataUrl, index_filename=indexFile)
            try:
                self._populateFromVariantFile(varFile, dataUrl, indexFile)
            finally:
                varFile.close()

    def populateFromDirectory(self, vcfDirectory):
        """
        Populates this VariantSet by examing all the VCF files in the
        specified directory. This is mainly used for as a convenience
        for testing purposes.
        """
        pattern = os.path.join(vcfDirectory, "*.vcf.gz")
        dataFiles = []
        indexFiles = []
        for vcfFile in glob.glob(pattern):
            dataFiles.append(vcfFile)
            indexFiles.append(vcfFile + ".tbi")
        self.populateFromFile(dataFiles, indexFiles)

    def getVcfHeaderReferenceSetName(self):
        """
        Returns the name of the reference set from the VCF header.
        """
        # TODO implemenent
        return None

    def checkConsistency(self):
        """
        Perform consistency check on the variant set
        """
        for referenceName, (dataUrl, indexFile) in self._chromFileMap.items():
            varFile = pysam.VariantFile(dataUrl, index_filename=indexFile)
            try:
                for chrom in varFile.index:
                    chrom, _, _ = self.sanitizeVariantFileFetch(chrom)
                    if not isEmptyIter(varFile.fetch(chrom)):
                        self._checkMetadata(varFile)
                        self._checkCallSetIds(varFile)
            finally:
                varFile.close()

    def _populateFromVariantFile(self, varFile, dataUrl, indexFile):
        """
        Populates the instance variables of this VariantSet from the specified
        pysam VariantFile object.
        """
        if varFile.index is None:
            raise exceptions.NotIndexedException(dataUrl)
        for chrom in varFile.index:
            # Unlike Tabix indices, CSI indices include all contigs defined
            # in the BCF header.  Thus we must test each one to see if
            # records exist or else they are likely to trigger spurious
            # overlapping errors.
            chrom, _, _ = self.sanitizeVariantFileFetch(chrom)
            if not isEmptyIter(varFile.fetch(chrom)):
                if chrom in self._chromFileMap:
                    raise exceptions.OverlappingVcfException(dataUrl, chrom)
            self._chromFileMap[chrom] = dataUrl, indexFile
        self._updateMetadata(varFile)
        self._updateCallSetIds(varFile)
        self._updateVariantAnnotationSets(varFile, dataUrl)

    def _updateVariantAnnotationSets(self, variantFile, dataUrl):
        """
        Updates the variant annotation set associated with this variant using
        information in the specified pysam variantFile.
        """
        # TODO check the consistency of this between VCF files.
        if not self.isAnnotated():
            annotationType = None
            for record in variantFile.header.records:
                if record.type == "GENERIC":
                    if record.key == "SnpEffVersion":
                        annotationType = ANNOTATIONS_SNPEFF
                    elif record.key == "VEP":
                        version = record.value.split()[0]
                        # TODO we need _much_ more sophisticated processing
                        # of VEP versions here. When do they become
                        # incompatible?
                        if version == "v82":
                            annotationType = ANNOTATIONS_VEP_V82
                        elif version == "v77":
                            annotationType = ANNOTATIONS_VEP_V77
                        else:
                            # TODO raise a proper typed exception there with
                            # the file name as an argument.
                            raise ValueError(
                                "Unsupported VEP version {} in '{}'".format(
                                    version, dataUrl))
            if annotationType is None:
                infoKeys = variantFile.header.info.keys()
                if 'CSQ' in infoKeys or 'ANN' in infoKeys:
                    # TODO likewise, we want a properly typed exception that
                    # we can throw back to the repo manager UI and display
                    # as an import error.
                    raise ValueError(
                        "Unsupported annotations in '{}'".format(dataUrl))
            if annotationType is not None:
                vas = HtslibVariantAnnotationSet(self, self.getLocalId())
                vas.populateFromFile(variantFile, annotationType)
                self.addVariantAnnotationSet(vas)

    def _updateMetadata(self, variantFile):
        """
        Updates the metadata for his variant set based on the specified
        variant file
        """
        metadata = self._getMetadataFromVcf(variantFile)
        if self._metadata is None:
            self._metadata = metadata

    def _checkMetadata(self, variantFile):
        """
        Checks that metadata is consistent
        """
        metadata = self._getMetadataFromVcf(variantFile)
        if self._metadata is not None and self._metadata != metadata:
            raise exceptions.InconsistentMetaDataException(
                variantFile.filename)

    def _checkCallSetIds(self, variantFile):
        """
        Checks callSetIds for consistency
        """
        if len(self._callSetIdMap) > 0:
            callSetIds = set([
                self.getCallSetId(sample)
                for sample in variantFile.header.samples])
            if callSetIds != set(self._callSetIdMap.keys()):
                raise exceptions.InconsistentCallSetIdException(
                    variantFile.filename)

    def getNumVariants(self):
        """
        Returns the total number of variants in this VariantSet.
        """
        # TODO How do we get the number of records in a VariantFile?
        return 0

    def _updateCallSetIds(self, variantFile):
        """
        Updates the call set IDs based on the specified variant file.
        """
        if len(self._callSetIdMap) == 0:
            for sample in variantFile.header.samples:
                self.addCallSetFromName(sample)

    def openFile(self, dataUrlIndexFilePair):
        dataUrl, indexFile = dataUrlIndexFilePair
        return pysam.VariantFile(dataUrl, index_filename=indexFile)

    def _convertGaCall(self, callSet, pysamCall):
        phaseset = None
        if pysamCall.phased:
            phaseset = str(pysamCall.phased)
        genotypeLikelihood = []
        info = {}
        for key, value in pysamCall.iteritems():
            if key == 'GL' and value is not None:
                genotypeLikelihood = list(value)
            elif key != 'GT':
                info[key] = protocol.encodeValue(value)
        call = protocol.Call()
        call.call_set_name = callSet.getSampleName()
        call.call_set_id = callSet.getId()
        call.genotype.extend(list(pysamCall.allele_indices))
        call.phaseset = pb.string(phaseset)
        call.genotype_likelihood.extend(genotypeLikelihood)
        for key in info:
            call.attributes.attr[key].values.extend(info[key])
        return call

    def convertVariant(self, record, callSetIds):
        """
        Converts the specified pysam variant record into a GA4GH Variant
        object. Only calls for the specified list of callSetIds will
        be included.
        """
        variant = self._createGaVariant()
        variant.reference_name = record.contig
        if record.id is not None:
            variant.names.extend(record.id.split(';'))
        variant.start = record.start          # 0-based inclusive
        variant.end = record.stop             # 0-based exclusive
        variant.reference_bases = record.ref
        if record.alts is not None:
            variant.alternate_bases.extend(list(record.alts))
        filterKeys = record.filter.keys()
        if len(filterKeys) == 0:
            variant.filters_applied = False
        else:
            variant.filters_applied = True
            if len(filterKeys) == 1 and filterKeys[0] == 'PASS':
                variant.filters_passed = True
            else:
                variant.filters_passed = False
                variant.filters_failed.extend(filterKeys)
        # record.qual is also available, when supported by GAVariant.
        for key, value in record.info.iteritems():
            if value is None:
                continue
            if key == 'SVTYPE':
                variant.variant_type = value
            elif key == 'SVLEN':
                variant.svlen = int(value[0])
            elif key == 'CIPOS':
                variant.cipos.extend(value)
            elif key == 'CIEND':
                variant.ciend.extend(value)
            elif isinstance(value, str):
                value = value.split(',')
            protocol.setAttribute(
                variant.attributes.attr[key].values, value)
        for callSetId in callSetIds:
            callSet = self.getCallSet(callSetId)
            pysamCall = record.samples[str(callSet.getSampleName())]
            variant.calls.add().CopyFrom(
                self._convertGaCall(callSet, pysamCall))
        variant.id = self.getVariantId(variant)
        return variant

    def getVariant(self, compoundId):
        if compoundId.reference_name in self._chromFileMap:
            varFileName = self._chromFileMap[compoundId.reference_name]
        else:
            raise exceptions.ObjectNotFoundException(compoundId)
        start = int(compoundId.start)
        referenceName, startPosition, endPosition = \
            self.sanitizeVariantFileFetch(
                compoundId.reference_name, start, start + 1)
        cursor = self.getFileHandle(varFileName).fetch(
            referenceName, startPosition, endPosition)
        for record in cursor:
            variant = self.convertVariant(record, self._callSetIds)
            if (record.start == start and
                    compoundId.md5 == self.hashVariant(variant)):
                return variant
            elif record.start > start:
                raise exceptions.ObjectNotFoundException()
        raise exceptions.ObjectNotFoundException(compoundId)

    def getPysamVariants(self, referenceName, startPosition, endPosition):
        """
        Returns an iterator over the pysam VCF records corresponding to the
        specified query.
        """
        if referenceName in self._chromFileMap:
            varFileName = self._chromFileMap[referenceName]
            referenceName, startPosition, endPosition = \
                self.sanitizeVariantFileFetch(
                    referenceName, startPosition, endPosition)
            cursor = self.getFileHandle(varFileName).fetch(
                referenceName, startPosition, endPosition)
            for record in cursor:
                yield record

    def getVariants(self, referenceName, startPosition, endPosition,
                    callSetIds=[]):
        """
        Returns an iterator over the specified variants. The parameters
        correspond to the attributes of a GASearchVariantsRequest object.
        """
        if callSetIds is None:
            callSetIds = self._callSetIds
        else:
            for callSetId in callSetIds:
                if callSetId not in self._callSetIds:
                    raise exceptions.CallSetNotInVariantSetException(
                        callSetId, self.getId())
        for record in self.getPysamVariants(
                referenceName, startPosition, endPosition):
            yield self.convertVariant(record, callSetIds)

    def getMetadataId(self, metadata):
        """
        Returns the id of a metadata
        """
        return str(datamodel.VariantSetMetadataCompoundId(
            self.getCompoundId(), 'metadata:' + metadata.key))

    def _getMetadataFromVcf(self, varFile):
        # All the metadata is available via each varFile.header, including:
        #    records: header records
        #    version: VCF version
        #    samples -- not immediately needed
        #    contigs -- not immediately needed
        #    filters -- not immediately needed
        #    info
        #    formats

        def buildMetadata(
                key, type_="String", number="1", value="", id_="",
                description=""):  # All input are strings
            metadata = protocol.VariantSetMetadata()
            metadata.key = key
            metadata.value = value
            metadata.type = type_
            metadata.number = number
            metadata.description = description
            if id_ == '':
                id_ = self.getMetadataId(metadata)
            metadata.id = id_
            return metadata

        ret = []
        header = varFile.header
        ret.append(buildMetadata(key="version", value=header.version))
        formats = header.formats.items()
        infos = header.info.items()
        filters = header.filters.items()
        # TODO: currently ALT field is not implemented through pysam
        # NOTE: contigs field is different between vcf files,
        # so it's not included in metadata
        for prefix, content in [("FORMAT", formats), ("INFO", infos),
                                ("FILTER", filters)]:
            for contentKey, value in content:
                description = value.description.strip('"')
                key = "{0}.{1}".format(prefix, value.name)
                if prefix == "FILTER":
                    ret.append(buildMetadata(
                        key=key,
                        description=description))
                elif key != "FORMAT.GT":
                    ret.append(buildMetadata(
                        key=key, type_=value.type,
                        number="{}".format(value.number),
                        description=description))
        return ret

#############################################

# Variant Annotations.

#############################################


class AbstractVariantAnnotationSet(datamodel.DatamodelObject):
    """
    Class representing a variant annotation set derived from an
    annotated variant set.
    """
    compoundIdClass = datamodel.VariantAnnotationSetCompoundId

    def __init__(self, variantSet, localId):
        super(AbstractVariantAnnotationSet, self).__init__(variantSet, localId)
        self._variantSet = variantSet
        self._ontology = None
        self._analysis = None
        self._creationTime = ''
        self._updatedTime = ''

    def setOntology(self, ontology):
        """
        Sets the Ontology used in this VariantAnnotationSet to
        translate sequence ontology term names into IDs to the
        specified value.
        """
        self._ontology = ontology

    def getCreationTime(self):
        """
        Returns the creation time for this VariantAnnotationSet
        """
        return self._creationTime

    def getUpdatedTime(self):
        """
        Returns the update time for this VariantAnnotationSet
        """
        return self._updatedTime

    def getOntology(self):
        """
        Returns the ontology term map used in this VariantAnnotationSet.
        """
        return self._ontology

    def getAnalysis(self):
        """
        Returns the Analysis object associated with this VariantAnnotationSet.
        """
        return self._analysis

    def getVariantSet(self):
        """
        Returns the VariantSet that this VariantAnnotationSet refers to.
        """
        return self._variantSet

    def _createGaVariantAnnotation(self):
        """
        Convenience method to set the common fields in a GA VariantAnnotation
        object from this variant set.
        """
        ret = protocol.VariantAnnotation()
        ret.created = self._creationTime
        ret.variant_annotation_set_id = self.getId()
        return ret

    def _createGaTranscriptEffect(self):
        """
        Convenience method to set the common fields in a GA TranscriptEffect
        object.
        """
        ret = protocol.TranscriptEffect()
        return ret

    def _createGaAlleleLocation(self):
        """
        Convenience method to set the common fields in a AlleleLocation
        object.
        """
        ret = protocol.AlleleLocation()
        return ret

    def toProtocolElement(self):
        """
        Converts this VariantAnnotationSet into its GA4GH protocol equivalent.
        """
        protocolElement = protocol.VariantAnnotationSet()
        protocolElement.id = self.getId()
        protocolElement.variant_set_id = self._variantSet.getId()
        protocolElement.name = self.getLocalId()
        protocolElement.analysis.CopyFrom(self.getAnalysis())
        self.serializeAttributes(protocolElement)
        return protocolElement

    def getTranscriptEffectId(self, gaTranscriptEffect):
        effs = [eff.term for eff in gaTranscriptEffect.effects]
        return hashlib.md5(
            "{}\t{}\t{}\t{}".format(
                gaTranscriptEffect.alternate_bases,
                gaTranscriptEffect.feature_id,
                effs, gaTranscriptEffect.hgvs_annotation)
            ).hexdigest()

    def hashVariantAnnotation(cls, gaVariant, gaVariantAnnotation):
        """
        Produces an MD5 hash of the gaVariant and gaVariantAnnotation objects
        """
        treffs = [treff.id for treff in gaVariantAnnotation.transcript_effects]
        return hashlib.md5(
            "{}\t{}\t{}\t".format(
                gaVariant.reference_bases, tuple(gaVariant.alternate_bases),
                treffs)
            ).hexdigest()

    def getVariantAnnotationId(self, gaVariant, gaAnnotation):
        """
        Produces a stringified compoundId representing a variant
        annotation.
        :param gaVariant:   protocol.Variant
        :param gaAnnotation: protocol.VariantAnnotation
        :return:  compoundId String
        """
        md5 = self.hashVariantAnnotation(gaVariant, gaAnnotation)
        compoundId = datamodel.VariantAnnotationCompoundId(
            self.getCompoundId(), gaVariant.reference_name,
            str(gaVariant.start), md5)
        return str(compoundId)


class SimulatedVariantAnnotationSet(AbstractVariantAnnotationSet):
    """
    A variant annotation set that doesn't derive from a data store.
    Used mostly for testing.
    """
    def __init__(self, variantSet, localId, randomSeed):
        super(SimulatedVariantAnnotationSet, self).__init__(
            variantSet, localId)
        self._randomSeed = randomSeed
        self._analysis = self._createAnalysis()

    def _createAnalysis(self):
        analysis = protocol.Analysis()
        analysis.created = datetime.datetime.now().isoformat() + "Z"
        analysis.updated = datetime.datetime.now().isoformat() + "Z"
        analysis.software.append("software")
        analysis.name = "name"
        analysis.description = "description"
        analysis.id = str(datamodel.VariantAnnotationSetAnalysisCompoundId(
            self._compoundId, "analysis"))
        return analysis

    def getVariantAnnotation(self, variant, randomNumberGenerator):
        ann = self.generateVariantAnnotation(variant, randomNumberGenerator)
        return ann

    def getVariantAnnotations(self, referenceName, start, end):
        for variant in self._variantSet.getVariants(referenceName, start, end):
            yield variant, self.generateVariantAnnotation(variant)

    def generateVariantAnnotation(self, variant):
        """
        Generate a random variant annotation based on a given variant.
        This generator should be seeded with a value that is unique to the
        variant so that the same annotation will always be produced regardless
        of the order it is generated in.
        """
        # To make this reproducible, make a seed based on this
        # specific variant.
        seed = self._randomSeed + variant.start + variant.end
        randomNumberGenerator = random.Random()
        randomNumberGenerator.seed(seed)
        ann = protocol.VariantAnnotation()
        ann.variant_annotation_set_id = str(self.getCompoundId())
        ann.variant_id = variant.id
        ann.created = datetime.datetime.now().isoformat() + "Z"
        # make a transcript effect for each alternate base element
        # multiplied by a random integer (1,5)
        for base in variant.alternate_bases:
            ann.transcript_effects.add().CopyFrom(
                self.generateTranscriptEffect(
                    variant, ann, base, randomNumberGenerator))
        ann.id = self.getVariantAnnotationId(variant, ann)
        return ann

    def _addTranscriptEffectLocations(self, effect, ann, variant):
        # TODO Make these valid HGVS values
        effect.hgvs_annotation.genomic = str(variant.start)
        effect.hgvs_annotation.transcript = str(variant.start)
        effect.hgvs_annotation.protein = str(variant.start)
        effect.protein_location.start = variant.start
        effect.cds_location.start = variant.start
        effect.cdna_location.start = variant.start
        return effect

    def _getRandomOntologyTerm(self, randomNumberGenerator):
        # TODO more mock options from simulated seqOnt?
        ontologyTuples = [
            ("intron_variant", "SO:0001627", "LOW"),
            ("exon_variant", "SO:0001791", "LOW"),
            ("misense_variant", "SO:0001583", "MODERATE"),
            ("stop_gained", "SO:0001587", "HIGH")]
        term = protocol.OntologyTerm()
        picked = randomNumberGenerator.choice(ontologyTuples)
        term.term, term.term_id, impact = picked
        return term, impact

    def _addTranscriptEffectOntologyTerm(self, effect, randomNumberGenerator):
        term, impact = self._getRandomOntologyTerm(randomNumberGenerator)
        effect.effects.add().CopyFrom(term)
        protocol.setAttribute(effect.attributes.attr["impact"].values, impact)
        return effect

    def _generateAnalysisResult(self, effect, ann, randomNumberGenerator):
        # TODO make these sensible
        analysisResult = protocol.AnalysisResult()
        analysisResult.analysis_id = "analysisId"
        analysisResult.result = "result string"
        analysisResult.score = randomNumberGenerator.randint(0, 100)
        return analysisResult

    def _addAnalysisResult(self, effect, ann, randomNumberGenerator):
        effect.analysis_results.add().CopyFrom(
            self._generateAnalysisResult(
                effect, ann, randomNumberGenerator))
        return effect

    def generateTranscriptEffect(
            self, variant, ann, alts, randomNumberGenerator):
        effect = self._createGaTranscriptEffect()
        effect.alternate_bases = alts
        # TODO how to make these featureIds sensical?
        effect.feature_id = "E4TB33F"
        effect = self._addTranscriptEffectLocations(effect, ann, variant)
        effect = self._addTranscriptEffectOntologyTerm(
            effect, randomNumberGenerator)
        effect = self._addTranscriptEffectOntologyTerm(
            effect, randomNumberGenerator)
        effect.id = self.getTranscriptEffectId(effect)
        effect = self._addAnalysisResult(effect, ann, randomNumberGenerator)
        return effect


class HtslibVariantAnnotationSet(AbstractVariantAnnotationSet):
    """
    Class representing a single variant annotation derived from an
    annotated variant set.
    """
    CSQ_FIELDS = ("alternate_bases", "gene", "feature_id",
                  "featureType", "effects", "cdnaPos", "cdsPos",
                  "protPos", "aminos", "codons", "existingVar",
                  "distance", "strand", "sift", "polyPhen",
                  "motifName", "motifPos", "highInfPos",
                  "motifScoreChange")
    VEP_FIELDS = ("alternate_bases", "effects", "impact", "symbol",
                  "geneName", "featureType", "feature_id",
                  "trBiotype", "exon", "intron",
                  "hgvs_annotation.transcript",
                  "hgvs_annotation.protein", "cdnaPos", "cdsPos",
                  "protPos", "aminos", "codons", "existingVar",
                  "distance", "strand", "symbolSource", "hgncId",
                  "hgvsOffset")
    SNPEFF_FIELDS = ("alternate_bases", "effects", "impact",
                     "geneName", "geneId", "featureType",
                     "feature_id", "trBiotype", "rank",
                     "hgvs_annotation.transcript",
                     "hgvs_annotation.protein", "cdnaPos", "cdsPos",
                     "protPos", "distance", "errsWarns")
    EXCLUDED_FIELDS = ("effects", "geneName", "gene", "geneId", "featureType",
                       "trBiotype", "rank", "cdnaPos", "cdsPos",
                       "protPos", "distance", "exon", "intron",
                       "aminos", "codons", "existingVar", "distance",
                       "strand", "symbolSource", "hgncId",
                       "hgvsOffset")

    def __init__(self, variantSet, localId):
        super(HtslibVariantAnnotationSet, self).__init__(variantSet, localId)

    def populateFromFile(self, varFile, annotationType):
        self._annotationType = annotationType
        self._analysis = self._getAnnotationAnalysis(varFile)
        self._creationTime = self._analysis.created
        self._updatedTime = datetime.datetime.now().isoformat() + "Z"

    def populateFromRow(self, annotationSetRecord):
        """
        Populates this VariantAnnotationSet from the specified DB row.
        """
        self._annotationType = annotationSetRecord.annotationtype
        self._analysis = protocol.fromJson(
            annotationSetRecord.analysis, protocol.Analysis)
        self._creationTime = annotationSetRecord.created
        self._updatedTime = annotationSetRecord.updated
        self.setAttributesJson(annotationSetRecord.attributes)

    def getAnnotationType(self):
        """
        Returns the type of variant annotations, allowing us to determine
        how to interpret the annotations within the VCF file.
        """
        return self._annotationType

    def _getAnnotationAnalysis(self, varFile):
        """
        Assembles metadata within the VCF header into a GA4GH Analysis object.

        :return: protocol.Analysis
        """
        header = varFile.header
        analysis = protocol.Analysis()
        formats = header.formats.items()
        infos = header.info.items()
        filters = header.filters.items()
        for prefix, content in [("FORMAT", formats), ("INFO", infos),
                                ("FILTER", filters)]:
            for contentKey, value in content:
                key = "{0}.{1}".format(prefix, value.name)
                if key not in analysis.attributes.attr:
                    analysis.attributes.attr[key].Clear()
                if value.description is not None:
                    analysis.attributes.attr[
                        key].values.add().string_value = value.description
        analysis.created = self._creationTime
        analysis.updated = self._updatedTime
        for r in header.records:
            # Don't add a key to info if there's nothing in the value
            if r.value is not None:
                if r.key not in analysis.attributes.attr:
                    analysis.attributes.attr[r.key].Clear()
                analysis.attributes.attr[r.key] \
                    .values.add().string_value = str(r.value)
            if r.key == "created" or r.key == "fileDate":
                # TODO handle more date formats
                try:
                    if '-' in r.value:
                        fmtStr = "%Y-%m-%d"
                    else:
                        fmtStr = "%Y%m%d"
                    analysis.created = datetime.datetime.strptime(
                        r.value, fmtStr).isoformat() + "Z"
                except ValueError:
                    # is there a logger we should tell?
                    # print("INFO: Could not parse variant annotation time")
                    pass  # analysis.create_date_time remains datetime.now()
            if r.key == "software":
                analysis.software.append(r.value)
            if r.key == "name":
                analysis.name = r.value
            if r.key == "description":
                analysis.description = r.value
        analysis.id = str(datamodel.VariantAnnotationSetAnalysisCompoundId(
            self._compoundId, "analysis"))
        return analysis

    def getVariantAnnotations(self, referenceName, startPosition, endPosition):
        """
        Generator for iterating through variant annotations in this
        variant annotation set.
        :param referenceName:
        :param startPosition:
        :param endPosition:
        :return: generator of protocol.VariantAnnotation
        """
        variantIter = self._variantSet.getPysamVariants(
            referenceName, startPosition, endPosition)
        for record in variantIter:
            yield self.convertVariantAnnotation(record)

    def convertLocation(self, pos):
        """
        Accepts a position string (start/length) and returns
        a GA4GH AlleleLocation with populated fields.
        :param pos:
        :return: protocol.AlleleLocation
        """
        if isUnspecified(pos):
            return None
        coordLen = pos.split('/')
        if len(coordLen) > 1:
            allLoc = self._createGaAlleleLocation()
            allLoc.start = int(coordLen[0]) - 1
            return allLoc
        return None

    def convertLocationHgvsC(self, hgvsc):
        """
        Accepts an annotation in HGVS notation and returns
        an AlleleLocation with populated fields.
        :param hgvsc:
        :return:
        """
        if isUnspecified(hgvsc):
            return None
        match = re.match(".*c.(\d+)(\D+)>(\D+)", hgvsc)
        if match:
            pos = int(match.group(1))
            if pos > 0:
                allLoc = self._createGaAlleleLocation()
                allLoc.start = pos - 1
                allLoc.reference_sequence = match.group(2)
                allLoc.alternate_sequence = match.group(3)
                return allLoc
        return None

    def convertLocationHgvsP(self, hgvsp):
        """
        Accepts an annotation in HGVS notation and returns
        an AlleleLocation with populated fields.
        :param hgvsp:
        :return: protocol.AlleleLocation
        """
        if isUnspecified(hgvsp):
            return None
        match = re.match(".*p.(\D+)(\d+)(\D+)", hgvsp, flags=re.UNICODE)
        if match is not None:
            allLoc = self._createGaAlleleLocation()
            allLoc.reference_sequence = match.group(1)
            allLoc.start = int(match.group(2)) - 1
            allLoc.alternate_sequence = match.group(3)
            return allLoc
        return None

    def addCDSLocation(self, effect, cdnaPos):
        hgvsC = effect.hgvs_annotation.transcript
        allele_location = None
        if not isUnspecified(hgvsC):
            allele_location = self.convertLocationHgvsC(hgvsC)
            if allele_location:
                effect.cds_location.CopyFrom(self.convertLocationHgvsC(hgvsC))
        if allele_location is None and self.convertLocation(cdnaPos):
                effect.cds_location.CopyFrom(self.convertLocation(cdnaPos))
        else:
            # These are not stored in the VCF
            effect.cds_location.alternate_sequence = ""
            effect.cds_location.reference_sequence = ""

    def addProteinLocation(self, effect, protPos):
        hgvsP = effect.hgvs_annotation.protein
        protein_location = None
        if not isUnspecified(hgvsP):
            protein_location = self.convertLocationHgvsP(hgvsP)
            if protein_location:
                effect.protein_location.CopyFrom(
                    self.convertLocationHgvsP(hgvsP))
        if protein_location is None and self.convertLocation(protPos):
            effect.protein_location.CopyFrom(self.convertLocation(protPos))

    def addCDNALocation(self, effect, cdnaPos):
        hgvsC = effect.hgvs_annotation.transcript
        if self.convertLocation(cdnaPos):
            effect.cdna_location.CopyFrom(self.convertLocation(cdnaPos))
        if self.convertLocationHgvsC(hgvsC):
            effect.cdna_location.alternate_sequence = \
                self.convertLocationHgvsC(hgvsC).alternate_sequence
            effect.cdna_location.reference_sequence = \
                self.convertLocationHgvsC(hgvsC).reference_sequence

    def addLocations(self, effect, protPos, cdnaPos):
        """
        Adds locations to a GA4GH transcript effect object
        by parsing HGVS annotation fields in concert with
        and supplied position values.
        :param effect: protocol.TranscriptEffect
        :param protPos: String representing protein position from VCF
        :param cdnaPos: String representing coding DNA location
        :return: effect protocol.TranscriptEffect
        """
        self.addCDSLocation(effect, cdnaPos)
        self.addCDNALocation(effect, cdnaPos)
        self.addProteinLocation(effect, protPos)
        return effect

    def convertTranscriptEffect(self, annStr, hgvsG):
        """
        Takes the ANN string of a SnpEff generated VCF, splits it
        and returns a populated GA4GH transcript effect object.
        :param annStr: String
        :param hgvsG: String
        :return: effect protocol.TranscriptEffect()
        """
        effect = self._createGaTranscriptEffect()
        effect.hgvs_annotation.CopyFrom(protocol.HGVSAnnotation())
        annDict = dict()
        if self._annotationType == ANNOTATIONS_SNPEFF:
            annDict = dict(zip(self. SNPEFF_FIELDS, annStr.split("|")))
        elif self._annotationType == ANNOTATIONS_VEP_V82:
            annDict = dict(zip(self.VEP_FIELDS, annStr.split("|")))
        else:
            annDict = dict(zip(self.CSQ_FIELDS, annStr.split("|")))
        annDict["hgvs_annotation.genomic"] = hgvsG if hgvsG else u''
        for key, val in annDict.items():
            try:
                protocol.deepSetAttr(effect, key, val)
            except AttributeError:
                if val and key not in self.EXCLUDED_FIELDS:
                    protocol.setAttribute(
                        effect.attributes.attr[key].values, val)
        effect.effects.extend(self.convertSeqOntology(annDict.get('effects')))
        self.addLocations(
            effect, annDict.get('protPos'), annDict.get('cdnaPos'))
        effect.id = self.getTranscriptEffectId(effect)
        return effect

    def convertSeqOntology(self, seqOntStr):
        """
        Splits a string of sequence ontology effects and creates
        an ontology term record for each, which are built into
        an array of return soTerms.
        :param seqOntStr:
        :return: [protocol.OntologyTerm]
        """
        return [
            self._ontology.getGaTermByName(soName)
            for soName in seqOntStr.split('&')]

    def convertVariantAnnotation(self, record):
        """
        Converts the specfied pysam variant record into a GA4GH variant
        annotation object using the specified function to convert the
        transcripts.
        """
        variant = self._variantSet.convertVariant(record, [])
        annotation = self._createGaVariantAnnotation()
        annotation.variant_id = variant.id
        gDots = record.info.get(b'HGVS.g')
        # Convert annotations from INFO field into TranscriptEffect
        transcriptEffects = []
        annotations = record.info.get(b'ANN') or record.info.get(b'CSQ')
        for i, ann in enumerate(annotations):
            hgvsG = gDots[i % len(variant.alternate_bases)] if gDots else None
            transcriptEffects.append(self.convertTranscriptEffect(ann, hgvsG))
        annotation.transcript_effects.extend(transcriptEffects)
        annotation.id = self.getVariantAnnotationId(variant, annotation)
        return variant, annotation



#!/usr/bin/env python
"""
Example of client query
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import ga4gh.client.client as client


def runDemo():

    httpClient = client.HttpClient("http://localhost:8000")
    iterator = httpClient.search_variants(
        "WyIxa2ctcDMtc3Vic2V0IiwidnMiLCJtdm5jYWxsIl0",
        reference_name="1", start=45000, end=50000)
    for variant in iterator:
        print(
            variant.reference_name, variant.start, variant.end,
            variant.reference_bases, variant.alternate_bases, sep="\t")


if __name__ == '__main__':
    runDemo()
